from typing import Iterable, Iterator, Optional, IO


_SECTION_SEPARATOR = "/"
_PATH_SEPARATOR = "."


def normalize(string: str, to_lower: bool = True) -> str:
    ret = ""
    for c in string:
        if c.isalnum() or c in ("-", "_"):
            ret += c
    return ret.lower() if to_lower else ret


def maybe_escape(value: str) -> str:
    if (
        # pylint: disable-next=too-many-boolean-expressions
        len(value) == 0
        or '"' in value
        or "\\" in value
        or "\n" in value
        or value[0].isspace()
        or value[-1].isspace()
    ):
        ret = ""
        for c in value:
            if c == '"':
                ret += r"\""
            elif c == "\n":
                ret += r"\n"
            elif c == "\\":
                ret += "\\\\"
            else:
                ret += c

        return '"' + ret + '"'

    return value


def unescape(value: str) -> str:
    if '"' not in value and "\\" not in value:
        return value

    if len(value) < 2:
        return value

    if value[0] == '"':
        value = value[1:]

    if value[-1] == '"':
        value = value[:-1]

    ret = ""
    escaped = False

    for c in value:
        if not escaped:
            if c == "\\":
                escaped = True
                continue

            ret += c

        else:
            if c == "n":
                ret += "\n"
            else:
                ret += c
            escaped = False

    return ret


def cleanup_string(string: str, /, sep: str, to_lower: bool) -> str:
    replacements = (
        (f" {sep}", f"{sep}"),
        (f"{sep} ", f"{sep}"),
        (f"{sep}{sep}", f"{sep}"),
    )

    for from_, to in replacements:
        while from_ in string:
            string = string.replace(from_, to)

    string = sep.join(
        normalize(fragment.strip(), to_lower) for fragment in string.split(sep)
    )

    return string.strip(sep)


def cleanup_section(section: str) -> str:
    return cleanup_string(section, sep=_SECTION_SEPARATOR, to_lower=True)


def cleanup_path(path: str) -> str:
    return cleanup_string(path, sep=_PATH_SEPARATOR, to_lower=False)


# TODO: turn into a function (like dump_to_ini maybe) that takes the data and
# yields lines of text, and then use file.writelines.
def save_to_file(
    file: IO[str],
    data: Iterable[tuple[str, Optional[str]]],
    *,
    blanklines: bool,
) -> None:
    need_space = False
    current_section = ""

    def extract_section(path: str) -> tuple[str, str]:
        if _SECTION_SEPARATOR not in path:
            return "", ""

        section, path = path.rsplit(_SECTION_SEPARATOR, 1)
        if _SECTION_SEPARATOR in section:
            _, section = section.split(_SECTION_SEPARATOR, 1)

        return section, path

    for path, dump in data:
        section, path = extract_section(path.strip())
        if not section:
            continue

        if section != current_section:
            current_section = section

            if need_space and blanklines:
                file.write("\n")
            need_space = True

            file.write(f"[{current_section}]\n")

        if path:
            file.write(f"{path} =")
            if dump is not None:
                file.write(f" {maybe_escape(dump)}")
            file.write("\n")


def load_from_file(
    file: IO[str], main: str
) -> Iterator[tuple[str, Optional[str]]]:
    main = normalize(main)

    current_section = ""

    for line in file:
        line = line.strip()

        if not line:
            continue

        if line[0] == "[" and line[-1] == "]":
            current_section = cleanup_section(line[1:-1])
            if not current_section:
                continue

            if current_section != main:
                current_section = main + _SECTION_SEPARATOR + current_section

            yield current_section + _SECTION_SEPARATOR, ""

        elif "=" in line:
            path, dump = line.split("=", 1)
            path = cleanup_path(path)
            dump = dump.strip()
            if path and current_section:
                payload = unescape(dump) if dump else None
                yield (current_section + _SECTION_SEPARATOR + path, payload)



import logging
import weakref

from types import GenericAlias
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    Optional,
    TypeVar,
    cast,
)

try:
    from typing import Self
except ImportError:
    # TODO: Remove once we deprecate support for Python 3.10.
    from typing_extensions import Self

from sunset.exporter import maybe_escape
from sunset.lockable import Lockable
from sunset.notifier import Notifier
from sunset.protocols import BaseField, Serializer, UpdateNotifier
from sunset.serializers import lookup


_T = TypeVar("_T")


class Key(Generic[_T], BaseField, Lockable):
    """
    A single setting key containing a typed value.

    Keys support inheritance. If a Key does not have a value explicitly set, and
    it has a parent, then its value will be that of its parent. Else its value
    if unset is the default value it was instantiated with.

    Keys can call a callback when their reported value changes, whether it's
    their own value that changed, or that inherited from a parent. Set this
    callback with the :meth:`onValueChangeCall()` method.

    You can control the values that can be set on this Key by passing a
    `validator` argument when instantiating it.

    Args:
        default: The value that this Key will return when not otherwise set; the
            type of this default determines the type of the values that can be
            set on this Key.

            If the type of the default is not one of bool, int, float, str, or
            an `enum.Enum` subclass, and is also not a class that implements the
            :class:`~sunset.Serializable` protocol, then a serializer argument
            must also be passed.

        serializer: An implementation of the :class:`~sunset.Serializer`
            protocol for the type of this Key's values. This argument must be
            passed if the type in question is not supported by a native
            SunsetSettings serializer. If a serializer is passed, it will be
            used even if SunsetSettings has its own serializer for that type.

        validator: A function that returns True if the given value can be set on
            this Key, else False. This allows you to control what values are
            allowable for this Key. Default: None.

        value_type: If given, the type that will be used by runtime safety
            checks instead of the type of the default value. This is rarely
            needed, but can be useful e.g. if the Key is meant to hold values
            from multiple possible subclasses of a base class.

    Example:

    >>> from sunset import Key
    >>> key: Key[int] = Key(default=0)
    >>> key.get()
    0
    >>> key.set(42)
    True
    >>> key.get()
    42
    >>> child_key: Key[int] = Key(default=0)
    >>> child_key.setParent(key)
    >>> child_key.get()
    42
    >>> child_key.set(101)
    True
    >>> child_key.get()
    101
    >>> key.set(36)
    True
    >>> key.get()
    36
    >>> child_key.get()
    101
    >>> child_key.clear()
    >>> child_key.get()
    36
    """

    _default: _T
    _value: Optional[_T]
    _serializer: Serializer[_T]
    _validator: Callable[[_T], bool]
    _bad_value_string: Optional[str]
    _value_change_notifier: Notifier[_T]
    _update_notifier: Notifier[[UpdateNotifier]]
    _loaded_notifier: Notifier[[]]
    _parent_ref: Optional[weakref.ref["Key[_T]"]]
    _children_ref: weakref.WeakSet["Key[_T]"]
    _type: type[_T]

    def __init__(
        self,
        default: _T,
        serializer: Optional[Serializer[_T]] = None,
        validator: Optional[Callable[[_T], bool]] = None,
        value_type: Optional[type[_T]] = None,
    ) -> None:
        super().__init__()

        # Keep a runtime reference to the practical type contained in this
        # key.

        if value_type is not None:
            if not isinstance(default, value_type):
                raise TypeError(
                    f"Default Key value '{default}' has type"
                    f" '{default.__class__.__name__}', which is not compatible"
                    " with explicitly given value type"
                    f" '{value_type.__name__}'."
                )
            self._type = value_type
        else:
            self._type = default.__class__

        self._default = default
        self._value = None
        self._bad_value_string = None

        if serializer is None:
            serializer = lookup(self._type)
            if serializer is None:
                raise TypeError(
                    f"Default Key value '{default}' has type"
                    f" '{self._type.__name__}', which is not"
                    " supported by a native serializer. Please construct"
                    " the Key with an explicit serializer argument."
                )

        if validator is not None:
            self._validator = validator
        else:
            self._validator = lambda _: True

        self._serializer = serializer

        self._value_change_notifier = Notifier()
        self._update_notifier = Notifier()
        self._loaded_notifier = Notifier()

        self._parent_ref = None
        self._children_ref = weakref.WeakSet()

    def get(self) -> _T:
        """
        Returns the current value of this Key.

        If this Key does not currently have a value set on it, return its
        fallback value, that being, the value of its parent if it has one, else
        the default.

        Returns:
            This Key's apparent value.
        """

        return self.fallback() if (value := self._value) is None else value

    def fallback(self) -> _T:
        """
        Returns the value that this Key will fall back to when it does not have
        a value currently set.

        That value is the value of its parent if it has one, or the default
        value for the Key if not.
        """

        return (
            self._default if (parent := self.parent()) is None else parent.get()
        )

    def set(self, value: _T) -> bool:
        """
        Sets the given value on this Key.

        Args:
            value: The value that this Key will now hold. Must be of the type
                bound to this Key, i.e. the same type as this Key's default
                value.

        Returns:
            True if the value was successfully set, else False, for instance if
            the validator refused the value.
        """

        # Safety check in case the user is holding it wrong.

        if not isinstance(value, self._type):
            return False

        if not self._validator(value):
            logging.debug(
                "Validator rejected value for Key %r: %r", self, value
            )
            return False

        # Setting a Key's value programmatically always resets bad values.

        self._bad_value_string = None

        previously_set = self.isSet()
        prev_value = self.get()

        self._value = value

        if prev_value != self.get():
            self._value_change_notifier.trigger(self.get())
            for child in self.children():
                child._notifyParentValueChanged()

        if not previously_set or prev_value != self.get():
            self._update_notifier.trigger(self)

        return True

    def clear(self) -> None:
        """
        Clears the value currently set on this Key, if any.
        """

        # Clearing the Key always resets bad values.

        self._bad_value_string = None

        if not self.isSet():
            return

        prev_value = self.get()
        self._value = None

        if prev_value != self.get():
            self._value_change_notifier.trigger(self.get())
            for child in self.children():
                child._notifyParentValueChanged()

        self._update_notifier.trigger(self)

    @Lockable.with_lock
    def updateValue(self, updater: Callable[[_T], _T]) -> None:
        """
        Atomically updates this Key's value using the given update function. The
        function will be called with the Key's current value, and the value it
        returns will be set as the Key's new value.

        Args:
            updater: A function that takes an argument of the same type held in
                this key, and returns an argument of the same type.
        """

        self.set(updater(self.get()))

    def isSet(self) -> bool:
        """
        Returns whether there is a value currently set on this Key.

        Returns:
            True if a value is set on this Key, else False.
        """

        return self._value is not None

    def onValueChangeCall(self, callback: Callable[[_T], Any]) -> None:
        """
        Adds a callback to be called whenever the value returned by calling
        :meth:`get()` on this Key would change, even if this Key itself was not
        updated; for instance, this will happen if there is no value currently
        set on it and its parent's value changed.

        The callback will be called with the new value as its argument.

        If you want a callback to be called whenever this Key is updated, even
        if its apparent value does not change, use :meth:`onUpdateCall()`
        instead. For example, if you call :meth:`set()` with a value of `0` on a
        Key newly created with a default value of `0`, callbacks added with
        :meth:`onUpdateCall()` are called and callbacks added with
        :meth:`onValueChangeCall()` are not.


        Args:
            callback: A callable that takes one argument of the same type as the
                values held by this Key.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._value_change_notifier.add(callback)

    def onUpdateCall(self, callback: Callable[["Key[_T]"], Any]) -> None:
        """
        Adds a callback to be called whenever this Key is updated, even if the
        value returned by :meth:`get()` does not end up changing.

        The callback will be called with this Key instance as its argument.

        If you want a callback to be called only when the apparent value of this
        Key changes, use :meth:`onValueChangeCall()` instead. For example, if a
        Key has no value set on it and has a parent whose value is updated, then
        callbacks added on this Key with :meth:`onValueChangeCall()` are called,
        and callbacks added with :meth:`onUpdateCall()` are not, because it's
        not *this* Key that was updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notifier.add(callback)  # type: ignore

    def onLoadedCall(self, callback: Callable[[], Any]) -> None:
        """
        Adds a callback to be called whenever settings were just loaded. This
        Key itself may or may not have been modified during the load.

        Args:
            callback: A callable that takes no argument.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._loaded_notifier.add(callback)

    def setValidator(self, validator: Callable[[_T], bool]) -> None:
        """
        Replaces this Key's validator.

        Args:
            validator: A function that returns True if the given value can be
                set on this Key, else False. This allows you to control what
                values are allowable for this Key.
        """

        self._validator = validator

    def setParent(self, parent: Optional[Self]) -> None:
        """
        Makes the given Key the parent of this one. If None, remove this
        Key's parent, if any.

        A Key with a parent will inherit its parent's value when this
        Key's own value is not currently set.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Key that will become this Key's parent, or
                None. The parent Key must have the same type as this
                Key.

        Note:
            A Key and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None:
            if type(self) is not type(parent):
                return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children_ref.discard(self)

        if parent is None:
            self._parent_ref = None
            return

        if parent._type is not self._type:
            # This should not happen... unless the user is holding it wrong.
            # So, better safe than sorry.

            return

        parent._children_ref.add(self)
        self._parent_ref = weakref.ref(parent)

    def parent(self) -> Optional[Self]:
        """
        Returns the parent of this Key, if any.

        Returns:
            A Key instance of the same type as this one, or None.
        """

        # Make the type of self._parent_ref more specific for the purpose of
        # type checking.

        parent = cast(Optional[weakref.ref[Self]], self._parent_ref)
        return parent() if parent is not None else None

    def children(self) -> Iterator[Self]:
        """
        Returns an iterator over the Keys that have this Key as their parent.

        Returns:
            An iterator over Keys of the same type as this one.
        """

        for child in self._children_ref:
            yield cast(Self, child)

    def dumpFields(self) -> Iterator[tuple[str, Optional[str]]]:
        if not self.skipOnSave():
            if self.isSet():
                yield "", self._serializer.toStr(self.get())

            elif self._bad_value_string is not None:
                # If a bad value was set in the settings file for this Key, and
                # the Key was not modified since, then save the bad value again.
                # This way, typos in the settings file don't outright destroy
                # the entry.

                yield "", self._bad_value_string

    def restoreField(self, path: str, value: Optional[str]) -> bool:
        if value is None:
            # Note that doing nothing when the given value is None, is
            # considered a success.

            return True

        if path != "":
            # Keys don't have sub-fields. If a path was given, then the path is
            # incorrect.
            return False

        with self._update_notifier.inhibit():

            if (val := self._serializer.fromStr(value)) is not None:
                return self.set(val)

            else:
                # Keep track of the value that failed to restore, so that we can
                # dump it again when saving. That way, if a user makes a typo
                # while editing the settings file, the faulty entry is not
                # entirely lost when we save.

                logging.error("Invalid value for Key %r: %s", self, value)
                self._bad_value_string = value
                return False

    def _notifyParentValueChanged(self) -> None:
        if self.isSet():
            return

        self._value_change_notifier.trigger(self.get())
        for child in self.children():
            child._notifyParentValueChanged()

    def _typeHint(self) -> GenericAlias:
        return GenericAlias(type(self), self._type)

    def _newInstance(self) -> Self:
        """
        Internal. Returns a new instance of this Key with the same default
        value.

        Returns:
            A new Key.
        """

        return self.__class__(
            default=self._default,
            serializer=self._serializer,
            validator=self._validator,
            value_type=self._type,
        )

    def __repr__(self) -> str:
        type_name = self._type.__name__
        value = maybe_escape(self._serializer.toStr(self.get()))
        if not self.isSet():
            value = f"({value})"
        return f"<Key[{type_name}]:{value}>"



import dataclasses
import inspect
import weakref

from typing import (
    Any,
    Callable,
    Iterator,
    MutableSet,
    Optional,
    cast,
)

try:
    from typing import Self
except ImportError:
    # TODO: Remove once we deprecate support for Python 3.10.
    from typing_extensions import Self


from sunset.notifier import Notifier
from sunset.protocols import (
    BaseField,
    Field,
    ItemTemplate,
    UpdateNotifier,
)
from sunset.sets import WeakNonHashableSet


class Bunch(BaseField):
    """
    A collection of related Keys.

    Under the hood, a Bunch is a dataclass, and can be used in the same manner,
    i.e. by defining attributes directly on the class itself.

    Example:

    >>> from sunset import Bunch, Key
    >>> class Appearance(Bunch):
    ...     class Font(Bunch):
    ...         name: Key[str] = Key(default="Arial")
    ...         size: Key[int] = Key(default=14)
    ...     main_font: Font      = Font()
    ...     secondary_font: Font = Font()
    >>> appearance = Appearance()
    >>> appearance.main_font.name.get()
    'Arial'
    >>> appearance.secondary_font.name.get()
    'Arial'
    >>> appearance.main_font.name.set("Times New Roman")
    True
    >>> appearance.secondary_font.name.set("Calibri")
    True
    >>> appearance.main_font.name.get()
    'Times New Roman'
    >>> appearance.secondary_font.name.get()
    'Calibri'
    """

    _parent_ref: Optional[weakref.ref["Bunch"]]
    _children_set: MutableSet["Bunch"]
    _fields: dict[str, Field]
    _update_notifier: Notifier[[UpdateNotifier]]
    _loaded_notifier: Notifier[[]]

    def __new__(cls) -> Self:
        # Build and return a dataclass constructed from this class. Keep a
        # reference to that dataclass as a private class attribute, so that we
        # only construct it once. This allows type identity checks (as in
        # "type(a) is type (b)") to work.

        _dataclass_attr = "__DATACLASS_CLASS"

        dataclass_class: Optional[type[Self]] = None
        if (dataclass_class := getattr(cls, _dataclass_attr, None)) is None:
            # We haven't yet constructed a dataclass from this class. Construct
            # one here.

            cls_parents = cls.__bases__

            dataclasses_fields: list[Any] = []

            potential_fields = list(vars(cls).items())
            for name, attr in potential_fields:
                if inspect.isclass(attr):
                    if attr.__name__ == name:
                        # This is probably a class definition that just happens
                        # to be located inside the containing Bunch definition.
                        # This is fine.

                        continue

                    raise TypeError(
                        f"Field '{name}' in the definition of '{cls.__name__}'"
                        " is uninstantiated"
                    )

                if isinstance(attr, ItemTemplate):
                    # Safety check: make sure the user isn't accidentally
                    # overriding an existing attribute.

                    for cls_parent in cls_parents:
                        if getattr(cls_parent, name, None) is not None:
                            raise TypeError(
                                f"Field '{name}' in the definition of"
                                f" '{cls.__name__}' overrides attribute of the"
                                " same name declared in parent class"
                                f" '{cls_parent.__name__}'; consider renaming"
                                f" this field to '{name}_' for instance"
                            )

                    # Create a proper field from the attribute.

                    field = dataclasses.field(default_factory=attr._newInstance)
                    dataclasses_fields.append((name, attr._typeHint(), field))

                    # Note that we delete the attribute now that the field is
                    # created. This helps avoid a hard-to-debug problem if the
                    # user subclasses a Bunch with a custom __init__() that
                    # doesn't call super().__init__(). That would seem to work,
                    # but any updates made to attributes of that bunch would
                    # really be applied to the attribute of the Bunch *class*,
                    # not the instance, which would cause 'weird action at a
                    # distance' bugs. Deleting the original class attribute
                    # prevents this entirely.

                    delattr(cls, name)

            # Create a dataclass based on this class. Note that we will be
            # providing our own __init__() override below.

            dataclass_class = cast(
                type[Self],
                dataclasses.make_dataclass(
                    cls.__qualname__,
                    dataclasses_fields,
                    init=False,
                    bases=(cls,) + cls_parents,
                ),
            )

            # And store it on the class itself for later.

            setattr(cls, _dataclass_attr, dataclass_class)

        # And finally, return an instance of the dataclass. Phew.

        return super().__new__(dataclass_class)  # type: ignore

    def __init__(self) -> None:
        super().__init__()

        # Set up the Bunch fields.

        # First, look up internal dataclass attribute names.

        fields_attr: str = getattr(dataclasses, "_FIELDS")
        field_type: Any = getattr(dataclasses, "_FIELD")
        fields: dict[str, dataclasses.Field[Any]] = getattr(
            self, fields_attr, {}
        )

        # Then look up the fields stored in the dataclass.

        for field in fields.values():
            if getattr(field, "_field_type", None) is not field_type:
                continue

            if field.default_factory is dataclasses.MISSING:
                continue

            setattr(self, field.name, field.default_factory())

        self.__post_init__()

    def __post_init__(self) -> None:
        self._parent_ref = None
        self._children_set = WeakNonHashableSet[Bunch]()
        self._fields = {}
        self._update_notifier = Notifier()
        self._loaded_notifier = Notifier()

        for label, field in vars(self).items():
            if isinstance(field, Field):
                self._fields[label] = field
                field.meta().update(label=label, container=self)
                field._update_notifier.add(self._update_notifier.trigger)
                self._loaded_notifier.add(field._loaded_notifier.trigger)

    def setParent(self, parent: Optional[Self]) -> None:
        """
        Makes the given Bunch the parent of this one. If None, remove this
        Bunch's parent, if any.

        All the Key, List and Bunch fields defined on this Bunch instance
        will be recursively reparented to the corresponding Key / List / Bunch
        field on the given parent.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Bunch that will become this Bunch's parent, or
                None. The parent Bunch must have the same type as this
                Bunch.

        Note:
            A Bunch and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None:
            if type(self) is not type(parent):
                return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children_set.discard(self)

        if parent is None:
            self._parent_ref = None
        else:
            self._parent_ref = weakref.ref(parent)
            parent._children_set.add(self)

        for label, field in self._fields.items():
            if parent is None:
                field.setParent(None)
                continue

            parent_field = parent._fields.get(label)
            if parent_field is None:
                # This is a safety check, but it shouldn't happen. By
                # construction self should be of the same type as parent, so
                # they should have the same attributes.

                continue

            assert type(field) is type(parent_field)
            field.setParent(parent_field)

    def parent(self) -> Optional[Self]:
        """
        Returns the parent of this Bunch, if any.

        Returns:
            A Bunch instance of the same type as this one, or None.
        """

        # Make the type of self._parent_ref more specific for the purpose of
        # type checking.

        parent = cast(Optional[weakref.ref[Self]], self._parent_ref)
        return parent() if parent is not None else None

    def children(self) -> Iterator[Self]:
        """
        Returns an iterator over the Bunch instances that have this Bunch
        as their parent.

        Returns:
            An iterator over Bunch instances of the same type as this one.
        """

        # Note that we iterate on a copy so that this will not break if a
        # different thread updates the contents during the iteration.

        for child in list(self._children_set):
            yield cast(Self, child)

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this Bunch is updated. A Bunch
        is considered updated when any of its fields is updated.

        The callback will be called with as its argument whichever field was
        just updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notifier.add(callback)

    def onLoadedCall(self, callback: Callable[[], Any]) -> None:
        """
        Adds a callback to be called whenever settings were just loaded. This
        Bunch itself may or may not have been modified during the load.

        Args:
            callback: A callable that takes no argument.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._loaded_notifier.add(callback)

    def isSet(self) -> bool:
        """
        Indicates whether this Bunch holds any field that is set.

        Returns:
            True if any field set on this Bunch is set, else False.
        """

        return any(field.isSet() for field in self._fields.values())

    def dumpFields(self) -> Iterator[tuple[str, Optional[str]]]:
        """
        Internal.
        """

        sep = self._PATH_SEPARATOR
        if not self.skipOnSave():
            for label, field in sorted(self._fields.items()):
                yield from (
                    (label + sep + path if path else label, item)
                    for path, item in field.dumpFields()
                )

    def restoreField(self, path: str, value: Optional[str]) -> bool:
        """
        Internal.
        """

        with self._update_notifier.inhibit():
            if self._PATH_SEPARATOR in path:
                field_label, path = path.split(self._PATH_SEPARATOR, 1)
            else:
                field_label, path = path, ""

            if (field := self._fields.get(field_label)) is not None:
                return field.restoreField(path, value)

        return False

    def _typeHint(self) -> type:
        return type(self)

    def _newInstance(self) -> Self:
        """
        Internal. Returns a new instance of this Bunch with the same fields.

        Returns:
            A Bunch instance.
        """

        return self.__class__()



import io

from sunset import exporter


def test_escape() -> None:
    assert exporter.maybe_escape("") == '""'
    assert exporter.maybe_escape("123") == "123"
    assert exporter.maybe_escape("test test") == "test test"
    assert exporter.maybe_escape("   ") == '"   "'
    assert exporter.maybe_escape(" a") == '" a"'
    assert exporter.maybe_escape("a ") == '"a "'
    assert exporter.maybe_escape('"') == r'"\""'
    assert exporter.maybe_escape("test\ntest") == r'"test\ntest"'
    assert exporter.maybe_escape(r"test\test") == r'"test\\test"'


def test_unescape() -> None:
    assert exporter.unescape("") == ""
    assert exporter.unescape('""') == ""
    assert exporter.unescape('"') == '"'
    assert exporter.unescape('"test') == "test"
    assert exporter.unescape('test"') == "test"
    assert exporter.unescape("123") == "123"
    assert exporter.unescape("test \ttest") == "test \ttest"
    assert exporter.unescape('"   "') == "   "
    assert exporter.unescape('" a"') == " a"
    assert exporter.unescape('"a "') == "a "
    assert exporter.unescape(r'"\""') == '"'
    assert exporter.unescape(r'"test\ntest"') == "test\ntest"
    assert exporter.unescape(r'"test\\test"') == r"test\test"


def test_escape_reversible() -> None:
    for string in (
        "",
        "123",
        "test test",
        "   ",
        " a",
        "a ",
        '"',
        "test\ntest",
        r"test\test",
    ):
        assert exporter.unescape(exporter.maybe_escape(string)) == string


def test_save() -> None:
    input = [
        ("main/a", "1"),
        ("main/b.c", "test"),
        ("main/d.1.e", "test 2\ntest 2"),
        ("main/d.2.e", "  "),
        ("main/level1/b.c", "sub test"),
        ("main/level1/level2/a", "sub sub test"),
    ]

    file = io.StringIO()

    exporter.save_to_file(file, input, blanklines=True)
    assert (
        file.getvalue()
        == """\
[main]
a = 1
b.c = test
d.1.e = "test 2\\ntest 2"
d.2.e = "  "

[level1]
b.c = sub test

[level1/level2]
a = sub sub test
"""
    )


def test_load() -> None:
    _main = "main"

    input = io.StringIO(
        """\
[main]
a = 1
b.c = test
d.1.e = "test 2\\ntest 2"
d.2.e = "  "

[level1]
b.c = sub test

[level1/level2]
a = sub sub test

[empty]
"""
    )

    assert list(exporter.load_from_file(input, _main)) == [
        ("main/", ""),
        ("main/a", "1"),
        ("main/b.c", "test"),
        ("main/d.1.e", "test 2\ntest 2"),
        ("main/d.2.e", "  "),
        ("main/level1/", ""),
        ("main/level1/b.c", "sub test"),
        ("main/level1/level2/", ""),
        ("main/level1/level2/a", "sub sub test"),
        ("main/empty/", ""),
    ]


def test_section_cleanup() -> None:
    for input, expected in (
        ("", ""),
        (" // / ", ""),
        (" test 1 / test 2 ", "test1/test2"),
        ("test1///test2", "test1/test2"),
        ("/test", "test"),
        ("test/", "test"),
        ("  //   / I'snt This a Test? / // Yes! / // / ", "isntthisatest/yes"),
    ):
        assert exporter.cleanup_section(input) == expected


def test_path_cleanup() -> None:
    for input, expected in (
        ("", ""),
        (" .. . ", ""),
        (" test 1 . test 2 ", "test1.test2"),
        ("test1...test2", "test1.test2"),
        (".test", "test"),
        ("test.", "test"),
        ("  ..   . I'snt This a Test? . .. Yes! . .. . ", "IsntThisaTest.Yes"),
    ):
        assert exporter.cleanup_path(input) == expected



import weakref

from enum import Enum, auto
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    MutableSequence,
    Optional,
    SupportsIndex,
    TypeVar,
    Union,
    cast,
    overload,
)

try:
    from typing import Self
except ImportError:
    # TODO: Remove once we deprecate support for Python 3.10.
    from typing_extensions import Self

from sunset.bunch import Bunch
from sunset.key import Key
from sunset.notifier import Notifier
from sunset.protocols import BaseField, UpdateNotifier
from sunset.sets import WeakNonHashableSet

ListItemT = TypeVar("ListItemT", bound=Union[Bunch, Key[Any]])


class IterOrder(Enum):
    NO_PARENT = auto()
    PARENT_FIRST = auto()
    PARENT_LAST = auto()


class List(MutableSequence[ListItemT], BaseField):
    """
    A list-like container for Keys or Bunches of a given type, to be used in a
    Settings' definition.

    It is type-compatible with standard Python lists and supports indexing,
    insertion, appending, etc.

    In addition, it offers support for update notification callbacks, and
    inheritance. The inheritance is used when iterating on the List's contents
    using the :meth:`iter()` method, which iterates on a List and, optionally,
    its parents.

    Instead of creating a new instance of the contained type and inserting or
    appending it, you can use the :meth:`appendOne()` or :meth:`insertOne()`
    methods to do it in one step.

    Args:
        template: A Key or a Bunch *instance* that represents the items that
            will be contained in this List. (The template itself will not be
            added to the List.)

        order: One of `List.NO_PARENT`, `List.PARENT_FIRST` or
            `List.PARENT_LAST`. Sets the default iteration order used in
            :meth:`iter()` when not otherwise specified. Default:
            `List.NO_PARENT`.

    Example:

    >>> from sunset import Bunch, Key, List, Settings
    >>> class ExampleSettings(Settings):
    ...     class ExampleBunch(Bunch):
    ...         a: Key[str] = Key(default="")
    ...     key_list: List[Key[int]]         = List(Key(default=0))
    ...     bunch_list: List[ExampleBunch] = List(ExampleBunch())
    >>> settings = ExampleSettings()
    >>> settings.bunch_list
    []
    >>> settings.bunch_list.appendOne().a.set("demo")
    True
    >>> settings.bunch_list
    [ExampleSettings.ExampleBunch(a=<Key[str]:demo>)]
    >>> settings.key_list
    []
    >>> settings.key_list.appendOne().set(12)
    True
    >>> settings.key_list
    [<Key[int]:12>]
    """

    NO_PARENT = IterOrder.NO_PARENT
    PARENT_FIRST = IterOrder.PARENT_FIRST
    PARENT_LAST = IterOrder.PARENT_LAST

    _contents: list[ListItemT]
    _parent_ref: Optional[weakref.ref["List[ListItemT]"]]
    _children_ref: WeakNonHashableSet["List[ListItemT]"]
    _iter_order: IterOrder
    _update_notifier: Notifier[[UpdateNotifier]]
    _loaded_notifier: Notifier[[]]
    _template: ListItemT

    def __init__(
        self, template: ListItemT, order: IterOrder = IterOrder.NO_PARENT
    ) -> None:
        super().__init__()

        self._contents = []

        self._parent_ref = None
        self._children_ref = WeakNonHashableSet()
        self._iter_order = order
        self._update_notifier = Notifier()
        self._loaded_notifier = Notifier()
        self._template = template

    def insert(self, index: SupportsIndex, value: ListItemT) -> None:
        self._contents.insert(index, value)
        self._relabelItems()
        self._update_notifier.trigger(self)

    @overload
    def __getitem__(self, index: SupportsIndex) -> ListItemT: ...

    @overload
    def __getitem__(self, index: slice) -> list[ListItemT]: ...

    def __getitem__(
        self, index: Union[SupportsIndex, slice]
    ) -> Union[ListItemT, list[ListItemT]]:
        return self._contents[index]

    @overload
    def __setitem__(self, index: SupportsIndex, value: ListItemT) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[ListItemT]) -> None: ...

    def __setitem__(
        self,
        index: Union[SupportsIndex, slice],
        value: Union[ListItemT, Iterable[ListItemT]],
    ) -> None:
        self._clearMetadata(self._contents[index])
        if isinstance(index, slice):
            assert isinstance(value, Iterable)
            self._contents[index] = value

        else:
            assert isinstance(index, SupportsIndex)
            assert not isinstance(value, Iterable)
            self._contents[index] = value

        self._relabelItems()
        self._update_notifier.trigger(self)

    def __delitem__(self, index: Union[SupportsIndex, slice]) -> None:
        self._clearMetadata(self._contents[index])
        del self._contents[index]
        self._relabelItems()
        self._update_notifier.trigger(self)

    def extend(self, values: Iterable[ListItemT]) -> None:
        self._contents.extend(values)
        self._relabelItems()
        self._update_notifier.trigger(self)

    def append(self, value: ListItemT) -> None:
        self.extend((value,))

    def __iadd__(self, values: Iterable[ListItemT]) -> Self:
        self.extend(values)
        return self

    def clear(self) -> None:
        del self[:]

    def _clearMetadata(self, fields: Union[ListItemT, list[ListItemT]]) -> None:
        for field in fields if isinstance(fields, list) else [fields]:
            field.meta().clear()
            field._update_notifier.discard(self._update_notifier.trigger)
            self._loaded_notifier.discard(field._loaded_notifier.trigger)

    def __len__(self) -> int:
        return len(self._contents)

    def _newItem(self) -> ListItemT:
        item = self._template._newInstance()

        # WORKAROUND: This lets mypy properly typecheck the return type.

        return cast(ListItemT, item)

    def isSet(self) -> bool:
        """
        Indicates whether this List holds any item that is set.

        Returns:
            True if any item contained in this List is set, else False.
        """

        return any(item.isSet() for item in self._contents)

    def appendOne(self) -> ListItemT:
        """
        Creates a new item of the type contained in this List, appends it to
        this List, and returns it.

        Returns:
            An instance of the item type contained in this List.
        """

        item = self._newItem()
        self.append(item)
        return item

    def insertOne(self, index: int) -> ListItemT:
        """
        Creates a new item of the type contained in this List, inserts it in
        this List at the given index, and returns it.

        Args:
            index: The list index where to insert the new item.

        Returns:
            An instance of the item type contained in this List.
        """

        item = self._newItem()
        self.insert(index, item)
        return item

    def _relabelItems(self) -> None:
        for i, item in enumerate(self._contents):
            item.meta().update(label=self._labelForIndex(i), container=self)
            item._update_notifier.add(self._update_notifier.trigger)
            self._loaded_notifier.add(item._loaded_notifier.trigger)

    @staticmethod
    def _labelForIndex(index: SupportsIndex) -> str:
        return str(int(index) + 1)

    @staticmethod
    def _indexForLabel(label: str) -> Optional[int]:
        if not label.isdigit():
            return None
        return int(label) - 1

    def setParent(self, parent: Optional[Self]) -> None:
        """
        Makes the given List the parent of this one. If None, remove this List's
        parent, if any.

        Having a parent does not affect a List's behavior outside of the
        :meth:`iter()` method.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a List that will become this List's parent, or
                None. The parent List must have the same type as this List.

        Note:
            A List and its parent, if any, do not increase each other's
            reference count.
        """

        # Runtime check to affirm the type check of the method.

        if parent is not None:
            if type(self) is not type(parent):
                return

        old_parent = self.parent()
        if old_parent is not None:
            old_parent._children_ref.discard(self)

        if parent is None:
            self._parent_ref = None
        else:
            self._parent_ref = weakref.ref(parent)
            parent._children_ref.add(self)

    def iter(self, order: Optional[IterOrder] = None) -> Iterator[ListItemT]:
        """
        Yields the elements contained in this List, and optionally in its
        parents, if any.

        Args:
            order: One of `List.NO_PARENT`, `List.PARENT_FIRST`,
                `List.PARENT_LAST`, or None.

                - If `List.NO_PARENT`, this method only yields the contents of
                  this List instance.

                - If `List.PARENT_FIRST`, it yields from this List's parents, if
                  any, then this List itself.

                - If `List.PARENT_LAST`, it yields from this List itself, then
                  from its parents, if any.

                - If None, the order set on the List itself at creation time
                  will be used.

                Default: None.

        Returns:
            An iterator over the items contained in this List and optionally its
            parents.

        Example:

        >>> from sunset import Key, List
        >>> show = lambda l: [key.get() for key in l]
        >>> parent = List(Key(default=0))
        >>> parent.appendOne().set(1)
        True
        >>> parent.appendOne().set(2)
        True
        >>> child = List(Key(default=0))
        >>> child.appendOne().set(3)
        True
        >>> child.appendOne().set(4)
        True
        >>> show(parent)
        [1, 2]
        >>> show(child)
        [3, 4]
        >>> child.setParent(parent)
        >>> show(child.iter())
        [3, 4]
        >>> show(child.iter(order=List.PARENT_FIRST))
        [1, 2, 3, 4]
        >>> show(child.iter(order=List.PARENT_LAST))
        [3, 4, 1, 2]
        """

        parent = self.parent()

        if order is None:
            order = self._iter_order

        if parent is not None and order == IterOrder.PARENT_FIRST:
            yield from parent.iter(order)

        yield from self._contents

        if parent is not None and order == IterOrder.PARENT_LAST:
            yield from parent.iter(order)

    def parent(self) -> Optional[Self]:
        """
        Returns the parent of this List, if any.

        Returns:
            A List instance of the same type as this one, or None.
        """

        # Make the type of self._parent_ref more specific for the purpose of
        # type checking.

        parent = cast(Optional[weakref.ref[Self]], self._parent_ref)
        return parent() if parent is not None else None

    def children(self) -> Iterator[Self]:
        """
        Returns an iterator over the List instances that have this List
        as their parent.

        Returns:
            An iterator over List instances of the same type as this one.
        """

        for child in self._children_ref:
            yield cast(Self, child)

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this List, *or* any item contained
        in this List, is updated.

        The callback will be called with whichever entity was updated as its
        argument: this List, or one of its items or sub-items.

        Adding new items to or deleting items from a List is considered an
        update of that List, not of the elements in question.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._update_notifier.add(callback)

    def onLoadedCall(self, callback: Callable[[], Any]) -> None:
        """
        Adds a callback to be called whenever settings were just loaded. This
        List itself may or may not have been modified during the load.

        Args:
            callback: A callable that takes no argument.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        self._loaded_notifier.add(callback)

    def dumpFields(self) -> Iterator[tuple[str, Optional[str]]]:
        """
        Internal.
        """

        sep = self._PATH_SEPARATOR
        if not self.skipOnSave():
            for i, item in enumerate(self._contents, start=1):
                label = str(i)
                if not item.isSet():
                    yield label, None
                else:
                    yield from (
                        (label + sep + path if path else label, child_item)
                        for path, child_item in item.dumpFields()
                    )

    def restoreField(self, path: str, value: Optional[str]) -> bool:
        """
        Internal.
        """

        if self._PATH_SEPARATOR in path:
            field_label, path = path.split(self._PATH_SEPARATOR, 1)
        else:
            field_label, path = path, ""

        with self._update_notifier.inhibit():
            index = self._indexForLabel(field_label)
            if index is not None and index >= 0:
                self._ensureMinimumLength(index + 1)
                return self[index].restoreField(path, value)

        return False

    def _ensureMinimumLength(self, length: int) -> None:
        missing_count = length - len(self)

        if missing_count > 0:
            self.extend((self._newItem() for _ in range(missing_count)))

    def _typeHint(self) -> GenericAlias:
        return GenericAlias(type(self), type(self._template))

    def _newInstance(self) -> Self:
        """
        Internal. Returns a new instance of this List capable of holding the
        same type.

        Returns:
            A new List.
        """

        return self.__class__(template=self._template, order=self._iter_order)

    def __repr__(self) -> str:
        parent = self.parent()
        items = [repr(item) for item in self.iter(order=IterOrder.NO_PARENT)]
        if parent is not None and self._iter_order == IterOrder.PARENT_FIRST:
            items.insert(0, "<parent>")
        if parent is not None and self._iter_order == IterOrder.PARENT_LAST:
            items.append("<parent>")

        ret = "["
        ret += ",".join(items)
        ret += "]"
        return ret



import logging
import os
import pathlib
import tempfile

from types import TracebackType

from typing import Any, Callable, IO, Optional, Protocol, TypeVar, Union

from sunset.timer import PersistentTimer, TimerProtocol

_TimerT = TypeVar("_TimerT", bound=TimerProtocol)


class _SavableProtocol(Protocol):
    """
    This protocol lets AutoSaver know how to use a Settings instance without
    having to import the actual Settings class.
    """

    def load(self, file: IO[str]) -> None: ...

    def save(self, file: IO[str], blanklines: bool = False) -> None: ...

    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None: ...


class AutoSaver:
    """
    AutoSaver is a helper class that can take care of loading and saving
    settings automatically and safely.

    On instantiation, it automatically loads settings from the given file path,
    if that file exists, unless the `load_on_init` argument is set to False.

    When saving, it uses a two-steps mechanism to ensure the atomicity of the
    operation. That is to say, the operation either entirely succeeds or
    entirely fails; AutoSaver will never write an incomplete settings file.

    Saving automatically creates the parent directories of the target path if
    those don't exist yet.

    If the `save_on_update` argument is set to True, AutoSaver will save the
    settings whenever they are updated. If the `save_on_delete` argument is set
    to True, AutoSaver will save the settings when its own instance is about to
    get deleted.

    AutoSaver can also be used as a context manager. Using it as a context
    manager is the recommended usage.

    Args:
        settings: The Settings instance to load to and save from.

        path: The full path to the file to load the settings from and save
            them to. If this file does not exist yet, it will be created when
            saving for the first time.

        save_on_update: Whether to save the settings when they
            are updated in any way. Default: True.

        save_on_delete: Whether to save the settings when this AutoSaver
            instance is deleted. Default: True.

        load_on_init: Whether to load the settings when instantiating this
            AutoSaver, provided the settings path exists. Default: True.

        save_delay: How long to wait, in seconds, before actually saving the
            settings when `save_on_update` is True and an update occurs. Setting
            this to a few seconds will batch updates for that long before
            triggering a save. If set to 0, the save is triggered immediately.
            Default: 0.

        logger: A logger instance that will be used to log OS errors, if any,
            while loading or saving settings. If none is given, the default root
            logger will be used.

    Example:

    >>> from sunset import AutoSaver, Settings
    >>> class ExampleSettings(Settings):
    ...     ...
    >>> settings = ExampleSettings()
    >>> with AutoSaver(settings, "~/.config/my_app.conf"):  # doctest: +SKIP
    ...     main_program_loop(settings)
    """

    _DIR_MODE: int = 0o755
    _FILE_MODE: int = 0o644
    _ENCODING: str = "UTF-8"

    _path: pathlib.Path
    _settings: _SavableProtocol
    _dirty: bool
    _save_on_update: bool
    _save_on_delete: bool
    _save_delay: float
    _save_timer: TimerProtocol
    _logger: logging.Logger

    def __init__(
        self,
        settings: _SavableProtocol,
        path: Union[pathlib.Path, str],
        *,
        save_on_update: bool = True,
        save_on_delete: bool = True,
        load_on_init: bool = True,
        save_delay: float = 0.0,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        if isinstance(path, str):
            path = pathlib.Path(path)

        path = path.expanduser()

        if logger is None:
            logger = logging.getLogger()

        self._dirty = False
        self._path = path
        self._save_on_update = save_on_update
        self._save_on_delete = save_on_delete
        self._save_delay = save_delay
        self._save_timer = PersistentTimer(self.saveIfNeeded)
        self._settings = settings
        self._settings.onUpdateCall(self._onSettingsUpdated)
        self._logger = logger

        if load_on_init:
            self.doLoad()

    def path(self) -> pathlib.Path:
        """
        Returns the settings path for this AutoSaver.

        Returns:
            A path. It may or may not exist; that's not up to AutoSaver.
        """

        return self._path

    def doLoad(self) -> bool:
        """
        Load the settings from this AutoSaver's settings file path, if it
        exists.

        Unsaved settings, if any, will be lost.

        Returns:
            True if loading succeeded, else False. If an unexpected error
            occurred, the error will be logged using the logger passed during
            initialization. Note that it's not considered an error if the
            settings file path does not exist yet.

        Note:
            Prefer letting AutoSaver load the settings during its initialization
            in order to avoid race conditions.
        """

        try:
            if (path := self._path).exists():
                self._logger.debug("Loading settings file '%s'...", path)
                with open(path, encoding=self._ENCODING) as f:
                    self._settings.load(f)
                self._logger.debug("Loaded.")
                return True

        except OSError as e:
            self._logger.error(
                "Error while loading from '%s': %s", self._path, e
            )

        return False

    def doSave(self) -> bool:
        """
        Unconditionally saves the settings attached to this AutoSaver instance.

        This method uses a two-steps mechanism to perform the save, in order to
        make the save atomic. The settings are first saved to a temporary file,
        and if successful, that temporary file then replaces the actual settings
        file.

        If the directory where the settings file is located does not exist, this
        method automatically creates it.

        Returns:
            True if saving succeeded, else False. If False, the error will be
            logged using the logger passed during initialization.
        """

        self._save_timer.cancel()

        directory = self._path.parent
        directory.mkdir(parents=True, mode=self._DIR_MODE, exist_ok=True)

        try:
            with tempfile.NamedTemporaryFile(
                dir=directory,
                prefix=self._path.name,
                mode="xt",
                encoding=self._ENCODING,
                delete=False,
            ) as tmp:
                self._logger.debug("Saving settings file '%s'...", self._path)
                self._settings.save(tmp.file, blanklines=True)
                os.chmod(tmp.name, self._FILE_MODE)
                os.rename(tmp.name, self._path)
                self._logger.debug("Saved.")

        except OSError as e:
            self._logger.error("Error while saving to '%s': %s", self._path, e)
            return False

        self._dirty = False
        return True

    def saveIfNeeded(self) -> bool:
        """
        Performs a save if and only if there are pending, unsaved changes in the
        settings attached to this AutoSaver instance.

        Returns:
            True if a save was performed and succeeded, else False.
        """

        save_needed = self._dirty

        self._save_timer.cancel()

        if save_needed:
            return self.doSave()

        return save_needed

    def _onSettingsUpdated(self, _: Any) -> None:
        self._dirty = True
        if self._save_on_update:
            self._save_timer.start(self._save_delay)

    def setSaveTimerClass(self, timer_class: type[_TimerT]) -> _TimerT:
        """
        Internal.
        """

        timer = timer_class(self.saveIfNeeded)
        self._save_timer = timer
        return timer

    def __del__(self) -> None:
        if self._save_on_delete:
            self.saveIfNeeded()

    def __enter__(self) -> "AutoSaver":
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.saveIfNeeded()



from typing import Optional

from sunset import serializers


class ExampleSerializable:
    def __init__(self, value: str) -> None:
        self._value = value

    def toStr(self) -> str:
        return self._value

    @staticmethod
    def fromStr(string: str) -> Optional["ExampleSerializable"]:
        return ExampleSerializable(string)


def test_serialize_int() -> None:
    serializer = serializers.lookup(int)
    assert serializer is not None
    assert serializer.toStr(42) == "42"


def test_serialize_str() -> None:
    serializer = serializers.lookup(str)
    assert serializer is not None
    assert serializer.toStr(" !! test !!") == " !! test !!"


def test_serialize_bool() -> None:
    serializer = serializers.lookup(bool)
    assert serializer is not None
    assert serializer.toStr(True) == "true"
    assert serializer.toStr(False) == "false"


def test_serialize_float() -> None:
    serializer = serializers.lookup(float)
    assert serializer is not None
    assert serializer.toStr(1.0) == "1.0"
    assert serializer.toStr(2.34e-56) == "2.34e-56"


def test_serialize_base_type_subclass() -> None:
    class TestInt(int):
        pass

    serializer = serializers.lookup(TestInt)
    assert serializer is not None
    assert serializer.toStr(TestInt(1)) == "1"


def test_serialize_serializable() -> None:
    t = ExampleSerializable("serializable")
    serializer = serializers.lookup(type(t))
    assert serializer is not None
    assert serializer.toStr(t) == "serializable"


def test_deserialize_str() -> None:
    serializer = serializers.lookup(str)
    assert serializer is not None
    assert serializer.fromStr("test") == "test"


def test_deserialize_int() -> None:
    serializer = serializers.lookup(int)
    assert serializer is not None
    assert serializer.fromStr("test") is None
    assert serializer.fromStr("   -32 ") == -32
    assert serializer.fromStr("00017") == 17


def test_deserialize_float() -> None:
    serializer = serializers.lookup(float)
    assert serializer is not None
    assert serializer.fromStr("test") is None
    assert serializer.fromStr("   -32.10 ") == -32.1
    assert serializer.fromStr("2.34e-56") == 2.34e-56


def test_deserialize_bool() -> None:
    serializer = serializers.lookup(bool)
    assert serializer is not None
    assert serializer.fromStr("YES")
    assert serializer.fromStr("  1 ")
    assert serializer.fromStr("true")

    assert not serializer.fromStr("NO")
    assert not serializer.fromStr("  0 ")
    assert not serializer.fromStr("false")

    assert serializer.fromStr("garbage") is None


def test_deserialize_base_type_subclass() -> None:
    class TestInt(int):
        pass

    serializer = serializers.lookup(TestInt)
    assert serializer is not None
    assert serializer.fromStr("1") == TestInt(1)
    assert type(serializer.fromStr("1")) is TestInt


def test_deserialize_serializable() -> None:
    serializer = serializers.lookup(ExampleSerializable)
    assert serializer is not None
    t = serializer.fromStr("this is a test")
    assert t is not None
    assert isinstance(t, ExampleSerializable)
    assert t.toStr() == "this is a test"


def test_serializer_not_found() -> None:
    assert serializers.lookup(object) is None



import io

from pytest_mock import MockerFixture

from sunset import Bunch, Key, List, Settings, normalize


def test_normalize() -> None:
    assert normalize("") == ""
    assert normalize("     ") == ""
    assert normalize("  A  B  ") == "ab"
    assert normalize("(a):?/b") == "ab"
    assert normalize("a-b_c") == "a-b_c"


class ExampleSettings(Settings):
    class ExampleBunch(Bunch):
        c = Key(default=0)
        d = Key(default=False)

    inner_bunch = ExampleBunch()
    bunch_list = List(ExampleBunch())
    key_list = List(Key(default=""))

    a = Key(default="")
    b = Key(default="")


class TestSettings:
    def test_new_named_section(self) -> None:
        settings = ExampleSettings()

        assert len(list(settings.sections())) == 0

        section = settings.newSection(name="same name")
        assert len(list(settings.sections())) == 1

        othersection = settings.getOrCreateSection(name="same name")
        assert len(list(settings.sections())) == 1

        assert othersection is section

    def test_dump_fields(self) -> None:
        # When no field is set, the name of the section should still be dumped.

        settings = ExampleSettings()
        assert list(settings.dumpFields()) == [
            ("main/", None),
        ]

        # This applies to subsections too.

        settings.newSection("empty")
        assert list(settings.dumpFields()) == [
            ("main/", None),
            ("main/empty/", None),
        ]

        # Set fields should be dumped, in alphabetic order.

        settings = ExampleSettings()
        settings.b.set("new b")
        settings.a.set("new a")
        settings.inner_bunch.c.set(40)
        settings.inner_bunch.d.set(True)
        settings.bunch_list.appendOne().c.set(100)
        settings.bunch_list.appendOne().d.set(True)
        settings.key_list.appendOne().set("one")
        settings.key_list.appendOne()
        settings.key_list.appendOne().set("three")
        assert list(settings.dumpFields()) == [
            ("main/a", "new a"),
            ("main/b", "new b"),
            ("main/bunch_list.1.c", "100"),
            ("main/bunch_list.2.d", "true"),
            ("main/inner_bunch.c", "40"),
            ("main/inner_bunch.d", "true"),
            ("main/key_list.1", "one"),
            ("main/key_list.2", None),
            ("main/key_list.3", "three"),
        ]

        # Anonymous sections, and subsections of anonymous sections, should not
        # be dumped.

        settings = ExampleSettings()
        settings.a.set("a")
        settings.bunch_list.appendOne().c.set(100)

        section1 = settings.newSection(name="Section 1")
        section1.a.set("sub a")
        section1.b.set("sub b")
        section1.bunch_list.appendOne().c.set(1000)

        section2 = settings.newSection(name="Section 2")
        section2.inner_bunch.d.set(False)

        anonymous = settings.newSection()
        anonymous.a.set("anonymous")

        subanonymous = anonymous.newSection(name="Should be ignored too")
        subanonymous.b.set("anonymous 2")

        subsection = section1.newSection(name="Subsection")
        subsection.inner_bunch.c.set(200)

        assert list(settings.dumpFields()) == [
            ("main/a", "a"),
            ("main/bunch_list.1.c", "100"),
            ("main/section1/a", "sub a"),
            ("main/section1/b", "sub b"),
            ("main/section1/bunch_list.1.c", "1000"),
            ("main/section1/subsection/inner_bunch.c", "200"),
            ("main/section2/inner_bunch.d", "false"),
        ]

        # Settings with a custom section name should use that name in dumps.

        settings.setSectionName("new")
        assert list(settings.dumpFields()) == [
            ("new/a", "a"),
            ("new/bunch_list.1.c", "100"),
            ("new/section1/a", "sub a"),
            ("new/section1/b", "sub b"),
            ("new/section1/bunch_list.1.c", "1000"),
            ("new/section1/subsection/inner_bunch.c", "200"),
            ("new/section2/inner_bunch.d", "false"),
        ]

        # Section should be dumped in alphabetic order.

        settings = ExampleSettings()
        settings.newSection("z")
        settings.newSection("mm")
        settings.newSection("aaa")
        assert list(settings.dumpFields()) == [
            ("main/", None),
            ("main/aaa/", None),
            ("main/mm/", None),
            ("main/z/", None),
        ]

    def test_restore_field(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        # Test restorting a field. As well, restoring a field should not trigger
        # a callback.

        settings = ExampleSettings()
        settings.onUpdateCall(callback)
        assert settings.a.get() == ""
        assert settings.restoreField("main/a", "test main a")
        assert settings.a.get() == "test main a"
        callback.assert_not_called()

        # Test restoring a field on an inner Bunch.

        assert settings.inner_bunch.c.get() == 0
        assert settings.restoreField("main/inner_bunch.c", "123")
        assert settings.inner_bunch.c.get() == 123
        callback.assert_not_called()

        # Test restoring a field on a subsection.

        assert settings.getSection("section1") is None
        assert settings.restoreField("main/section1/a", "test section1 a")
        section1 = settings.getSection("section1")
        assert section1 is not None
        assert section1.a.get() == "test section1 a"
        callback.assert_not_called()

        assert settings.restoreField(
            "main/section2/subsection/a", "test subsection a"
        )
        section2 = settings.getSection("section2")
        assert section2 is not None
        subsection = section2.getSection("subsection")
        assert subsection is not None
        assert subsection.a.get() == "test subsection a"
        callback.assert_not_called()

        # Test restoring fields on a renamed Settings.

        renamed_settings = ExampleSettings()
        renamed_settings.setSectionName("renamed")
        renamed_settings.onUpdateCall(callback)
        assert renamed_settings.restoreField("renamed/a", "test renamed a")
        assert renamed_settings.a.get() == "test renamed a"
        callback.assert_not_called()

        assert renamed_settings.b.get() == ""
        assert not renamed_settings.restoreField("main/b", "test invalid b")
        assert renamed_settings.b.get() == ""
        callback.assert_not_called()

        # Test restorting with an invalid path.

        other_settings = ExampleSettings()
        other_settings.onUpdateCall(callback)
        assert not other_settings.isSet()
        assert not other_settings.restoreField("invalid/a", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

        assert not other_settings.restoreField("main/invalid", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

        assert not other_settings.restoreField("invalid", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

        assert not other_settings.restoreField("/a", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

        assert not other_settings.restoreField("a/invalid", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

        assert not other_settings.restoreField("a", "test invalid a")
        assert not other_settings.isSet()
        callback.assert_not_called()

    def test_persistence(self) -> None:
        # Settings keep a reference to their sections, but not to their parent.

        settings = ExampleSettings()
        level1 = settings.newSection(name="level 1")
        level2 = level1.newSection(name="level 2")

        assert level1.parent() is not None
        assert len(list(level1.sections())) == 1

        del settings
        del level2
        assert level1.parent() is None
        assert len(list(level1.sections())) == 1

    def test_save(self) -> None:
        settings = ExampleSettings()
        settings.a.set("a")
        settings.bunch_list.appendOne().c.set(100)

        section1 = settings.newSection(name="Level 1")
        section1.a.set("sub a")
        section1.b.set("sub b")
        section1.bunch_list.appendOne().c.set(1000)
        section1.key_list.appendOne().set("one")
        section1.key_list.appendOne()
        section1.key_list.appendOne().set("")

        section2 = settings.newSection(name="Other level 1")
        section2.inner_bunch.d.set(False)

        anonymous = settings.newSection()
        anonymous.a.set("anonymous")

        subanonymous = anonymous.newSection(name="Should be ignored too")
        subanonymous.b.set("anonymous 2")

        subsection1 = section1.newSection(name="Level 2")
        subsection1.inner_bunch.c.set(200)

        settings.newSection("empty")

        file = io.StringIO()
        settings.save(file, blanklines=True)

        assert (
            file.getvalue()
            == """\
[main]
a = a
bunch_list.1.c = 100

[empty]

[level1]
a = sub a
b = sub b
bunch_list.1.c = 1000
key_list.1 = one
key_list.2 =
key_list.3 = ""

[level1/level2]
inner_bunch.c = 200

[otherlevel1]
inner_bunch.d = false
"""
        )

    def test_load(self, mocker: MockerFixture) -> None:
        settings = ExampleSettings()
        callback = mocker.stub()
        settings.onUpdateCall(callback)
        settings.load(
            io.StringIO(
                """\
[main]
a = a
bunch_list.1.c = 100

[level1]
a = sub a
b = sub b
key_list.1 = one
key_list.2 =
key_list.3 = ""
bunch_list.1.c = 1000

[level1/level2]
inner_bunch.c = 200

[otherlevel1]
inner_bunch.d = false
"""
            )
        )

        callback.assert_not_called()

        assert settings.a.get() == "a"
        assert len(settings.bunch_list) == 1
        assert settings.bunch_list[0].c.get() == 100

        settings_sections = {s.sectionName(): s for s in settings.sections()}
        assert len(settings_sections) == 2
        assert "level1" in settings_sections
        level1 = settings_sections["level1"]
        assert "otherlevel1" in settings_sections
        otherlevel1 = settings_sections["otherlevel1"]

        assert level1.a.get() == "sub a"
        assert level1.b.get() == "sub b"
        assert len(level1.bunch_list) == 1
        assert level1.bunch_list[0].c.get() == 1000
        assert len(level1.key_list) == 3
        assert level1.key_list[0].get() == "one"
        assert not level1.key_list[1].isSet()
        assert level1.key_list[2].isSet() and level1.key_list[2].get() == ""

        assert not otherlevel1.inner_bunch.d.get()

        level1_sections = {s.sectionName(): s for s in level1.sections()}
        assert len(level1_sections) == 1
        assert "level2" in level1_sections
        level2 = level1_sections["level2"]

        assert level2.inner_bunch.c.get() == 200

    def test_load_invalid_no_section(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
a = no bunch header
"""
            )
        )

        assert settings.a.get() == ""

    def test_load_invalid_repeated_key(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a = repeated key
a = last value should be used
"""
            )
        )

        assert settings.a.get() == "last value should be used"

    def test_load_invalid_repeated_section(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a = repeated bunch

[main]
a = bunch values will be merged, last takes precedence
"""
            )
        )

        assert (
            settings.a.get()
            == "bunch values will be merged, last takes precedence"
        )

    def test_load_invalid_missing_main(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[level1]
a = main bunch is implicitly created if needed
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 1
        assert (
            sections[0].a.get() == "main bunch is implicitly created if needed"
        )

    def test_load_invalid_extra_section_separators(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
[level1///level2]
a = extra separators should be skipped
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 1
        subsections = list(sections[0].sections())
        assert len(subsections) == 1
        assert subsections[0].a.get() == "extra separators should be skipped"

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
[/level1]
a = extra separators should be skipped
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 1
        assert sections[0].a.get() == "extra separators should be skipped"

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
[level1/]
a = extra separators should be skipped
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 1
        assert sections[0].a.get() == "extra separators should be skipped"

    def test_load_invalid_bad_section_is_skipped(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a = main

[!%$?]
a = bad bunch
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 0
        assert settings.a.get() == "main"

    def test_load_invalid_similar_are_merged(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a = main

[ M? a*i*n ]
a = merged
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 0
        assert settings.a.get() == "merged"

    def test_load_invalid_empty_section_is_skipped(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a = main

[]
a = skipped
"""
            )
        )

        sections = list(settings.sections())
        assert len(sections) == 0
        assert settings.a.get() == "main"

    def test_load_bad_key(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
= should be skipped
"""
            )
        )

        assert not settings.isSet()

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
??a = should be loaded
"""
            )
        )

        assert settings.a.get() == "should be loaded"

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
[a] = should be loaded
"""
            )
        )

        assert settings.a.get() == "should be loaded"

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
doesnotexist = should be skipped
"""
            )
        )

        assert not settings.isSet()

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
. = should be skipped
"""
            )
        )

        assert not settings.isSet()

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
.a = should be loaded
"""
            )
        )

        assert settings.a.get() == "should be loaded"

        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[main]
a.. = should be loaded
"""
            )
        )

        assert settings.a.get() == "should be loaded"

    def test_load_sections_created_even_if_empty(self) -> None:
        settings = ExampleSettings()
        settings.load(
            io.StringIO(
                """\
[shouldexist]
"""
            )
        )
        assert settings.getSection("shouldexist") is not None

    def test_load_renamed_settings(self) -> None:
        settings = ExampleSettings()
        settings.setSectionName("renamed")

        settings.load(
            io.StringIO(
                """\
[renamed]
a = value
"""
            )
        )

        assert settings.a.get() == "value"

    def test_autosave(self, mocker: MockerFixture) -> None:
        sentinel = object()
        stub = mocker.Mock(return_value=sentinel)

        settings = ExampleSettings()
        settings.setAutosaverClass(stub)
        ret = settings.autosave(
            "/tmp/test",
            save_delay=12,
        )

        assert ret is sentinel

        stub.assert_called_once_with(
            settings,
            "/tmp/test",
            save_delay=12,
            save_on_update=True,
            logger=None,
        )

    def test_callback_type_is_flexible(self) -> None:
        settings = ExampleSettings()

        class Dummy:
            pass

        def callback(_: ExampleSettings) -> Dummy:
            return Dummy()

        settings.onUpdateCall(callback)

    def test_key_updated_notification(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        settings = ExampleSettings()
        settings.onUpdateCall(callback)

        settings.a.set("test 1")
        callback.assert_called_once_with(settings.a)
        callback.reset_mock()

        section = settings.newSection(name="section")
        callback.assert_called_once_with(settings)
        callback.reset_mock()

        section.a.set("test 2")
        callback.assert_called_once_with(section.a)
        callback.reset_mock()

        anonymous = settings.newSection()
        callback.assert_called_once_with(settings)
        callback.reset_mock()

        anonymous.a.set("test 3")
        callback.assert_called_once_with(anonymous.a)
        callback.reset_mock()

        anonymoussection = anonymous.newSection(name="other section")
        callback.assert_called_once_with(anonymous)
        callback.reset_mock()

        anonymoussection.a.set("test 4")
        callback.assert_called_once_with(anonymoussection.a)
        callback.reset_mock()

        anonymous.setSectionName("no longer anonymous")
        callback.assert_called_once_with(anonymous)
        callback.reset_mock()

    def test_setting_same_section_name_doesnt_notify(
        self, mocker: MockerFixture
    ) -> None:
        class TestSettings(Settings):
            pass

        parent = TestSettings()

        section = parent.newSection("test")

        section.onUpdateCall(callback := mocker.stub())

        section.setSectionName("test")
        callback.assert_not_called()

        section.setSectionName("")
        callback.assert_called_once_with(section)
        callback.reset_mock()

        section.setSectionName("")
        callback.assert_not_called()

    def test_reparenting_notifications(self, mocker: MockerFixture) -> None:
        class TestSettings(Settings):
            pass

        parent1 = TestSettings()
        parent2 = TestSettings()

        parent1.onUpdateCall(callback1 := mocker.stub())
        parent2.onUpdateCall(callback2 := mocker.stub())

        section = TestSettings()

        section.setParent(parent1)
        callback1.assert_called_once_with(parent1)
        callback2.assert_not_called()

        callback1.reset_mock()
        callback2.reset_mock()

        section.setParent(parent2)
        callback1.assert_called_once_with(parent1)
        callback2.assert_called_once_with(parent2)

        callback1.reset_mock()
        callback2.reset_mock()

        # When there is no change, no notification should fire.

        section.setParent(parent2)
        callback1.assert_not_called()
        callback2.assert_not_called()

        callback1.reset_mock()
        callback2.reset_mock()

        section.setParent(None)
        callback1.assert_not_called()
        callback2.assert_called_once_with(parent2)

        callback1.reset_mock()
        callback2.reset_mock()

        # The section is now detached and its update should no longer trigger
        # callbacks on the previous parents.

        section.setSectionName("dummy")
        callback1.assert_not_called()
        callback2.assert_not_called()

    def test_reparenting_renaming_notifications(
        self, mocker: MockerFixture
    ) -> None:
        class TestSettings(Settings):
            pass

        parent = TestSettings()
        parent.newSection("test")
        parent.onUpdateCall(parent_callback := mocker.stub())

        section = TestSettings()
        section.onUpdateCall(section_callback := mocker.stub())

        section.setSectionName("test")
        parent_callback.assert_not_called()
        section_callback.assert_called_once_with(section)

        parent_callback.reset_mock()
        section_callback.reset_mock()

        # This call adds the section to the parent AND renames the section.
        # Therefore both callbacks should be invoked.

        section.setParent(parent)
        parent_callback.assert_called_once_with(parent)
        section_callback.assert_called_once_with(section)

    def test_section_name_unicity(self) -> None:
        class TestSettings(Settings):
            pass

        parent = TestSettings()
        sections = [parent.newSection() for _ in range(10)]
        for section in sections:
            section.setSectionName("test")

        assert len(list(parent.children())) == len(
            set(child.sectionName() for child in parent.children())
        )

    def test_section_name_made_unique_when_changing_parent(self) -> None:
        class TestSettings(Settings):
            pass

        parent = TestSettings()
        parent.newSection("test")

        section = TestSettings()
        section.setSectionName("test")

        assert section.sectionName() == "test"

        section.setParent(parent)

        assert section.sectionName() != "test"

    def test_anonymous_name_not_unique(self) -> None:
        class TestSettings(Settings):
            pass

        parent = TestSettings()
        sections = [parent.newSection() for _ in range(10)]
        for section in sections:
            section.setSectionName("")

        assert all(section.sectionName() == "" for section in sections)



import threading

from typing import Any, Callable, Optional, Protocol


class TimerProtocol(Protocol):
    def __init__(self, function: Callable[[], Any]) -> None: ...

    def start(self, interval: float) -> None: ...

    def cancel(self) -> None: ...


class PersistentTimer:
    _timer: Optional[threading.Timer]
    _function: Callable[[], Any]
    _lock: threading.Lock

    def __init__(self, function: Callable[[], Any]) -> None:
        self._timer = None
        self._function = function
        self._lock = threading.Lock()

    def start(self, interval: float) -> None:
        if interval <= 0.0:
            self._timeout()
            return

        with self._lock:
            if self._timer is None:
                self._timer = threading.Timer(interval, self._timeout)
                self._timer.start()

    def cancel(self) -> None:
        with self._lock:
            if self._timer is not None:
                self._timer.cancel()
                self._timer = None

    def _timeout(self) -> None:
        self.cancel()
        self._function()

    def __del__(self) -> None:
        self.cancel()



import io

from pytest_mock import MockerFixture

from sunset import Bunch, Key, List, Settings


class ExampleSettings(Settings):
    class ExampleBunch(Bunch):
        c = Key(default=0)
        d = Key(default=False)

    inner_bunch = ExampleBunch()
    bunch_list = List(ExampleBunch())
    key_list = List(Key(default=""))

    a = Key(default="")
    b = Key(default="")


def test_loaded_notification(mocker: MockerFixture) -> None:
    settings = ExampleSettings()
    settings.key_list.appendOne()

    settings.onLoadedCall(callback_settings := mocker.stub())
    settings.a.onLoadedCall(callback_a := mocker.stub())
    settings.key_list.onLoadedCall(callback_key_list := mocker.stub())
    settings.key_list[0].onLoadedCall(callback_key_list_0 := mocker.stub())
    settings.inner_bunch.onLoadedCall(callback_inner_bunch := mocker.stub())
    settings.inner_bunch.c.onLoadedCall(callback_inner_bunch_c := mocker.stub())

    callback_settings.assert_not_called()
    callback_a.assert_not_called()
    callback_key_list.assert_not_called()
    callback_key_list_0.assert_not_called()
    callback_inner_bunch.assert_not_called()
    callback_inner_bunch_c.assert_not_called()

    settings.load(io.StringIO())

    callback_settings.assert_called_once()
    callback_a.assert_called_once()
    callback_key_list.assert_called_once()
    callback_key_list_0.assert_called_once()
    callback_inner_bunch.assert_called_once()
    callback_inner_bunch_c.assert_called_once()



from sunset import Bunch, Key, List, Settings


class ExampleBunch(Bunch):
    bunch_key = Key(default=0)


class ExampleSettings(Settings):
    inner_bunch = ExampleBunch()
    bunch_list = List(ExampleBunch())
    key_list = List(Key(default=""))
    key = Key(default="")


class TestMetadata:
    def test_key_label(self) -> None:
        class TestBunch(Bunch):
            key1 = Key("test")
            key2 = Key("test")

        bunch = TestBunch()
        assert bunch.key1.meta().label == "key1"
        assert bunch.key2.meta().label == "key2"

        key = Key("test")
        assert key.meta().label == ""

    def test_bunch_field_label(self) -> None:
        bunch = ExampleBunch()

        assert bunch.meta().label == ""
        assert bunch.bunch_key.meta().label == "bunch_key"

        settings = ExampleSettings()

        assert settings.meta().label == ""
        assert settings.inner_bunch.meta().label == "inner_bunch"
        assert settings.inner_bunch.bunch_key.meta().label == "bunch_key"

    def test_label_set_on_contained_list_items(self) -> None:
        # Testing List.appendOne()

        key_list = List(Key(""))
        key = key_list.appendOne()
        assert key.meta().label == "1"

        # Testing List.insertOne()

        other_key = key_list.insertOne(0)
        assert other_key.meta().label == "1"
        assert key.meta().label == "2"

        # Testing List.append()

        key = Key("")
        assert key.meta().label == ""
        key_list.append(key)
        assert key.meta().label == "3"

        # Testing List.extend()

        key = Key("")
        last_key = Key("")
        key_list.extend((key, last_key))
        assert key.meta().label == "4"
        assert last_key.meta().label == "5"

        # Testing List.insert()

        key = Key("")
        key_list.insert(3, key)
        assert key.meta().label == "4"
        assert last_key.meta().label == "6"

        # Testing List.__setitem__(index)

        key = Key("")
        key_list[2] = key
        assert key.meta().label == "3"
        assert last_key.meta().label == "6"

        # Testing List.__setitem__(slice)

        key = Key("")
        other_key = Key("")
        key_list[1:4] = [key, other_key]
        assert key.meta().label == "2"
        assert other_key.meta().label == "3"
        assert last_key.meta().label == "5"

        # Testing List.__iadd__()

        key = Key("")
        other_key = Key("")
        key_list += [key, other_key]
        assert key.meta().label == "6"
        assert other_key.meta().label == "7"

        # Testing assignment order

        key_list[2], key_list[4] = key_list[4], key_list[2]
        assert key_list[2].meta().label == "3"
        assert key_list[4].meta().label == "5"

    def test_label_unset_on_removed_list_items(self) -> None:
        key_list = List(Key(""))
        for _ in range(15):
            key_list.appendOne()

        # Test List.__delitem__(index)

        key = key_list[3]
        assert key.meta().label == "4"
        del key_list[3]
        assert key.meta().label == ""

        # Test List.__delitem__(slice)

        keys = key_list[2:5]
        assert [key.meta().label for key in keys] == ["3", "4", "5"]
        del key_list[2:5]
        assert [key.meta().label for key in keys] == ["", "", ""]

        # Test List.__setitem__(index)

        key = key_list[5]
        assert key.meta().label == "6"
        key_list[5] = Key("")
        assert key.meta().label == ""

        # Test List.__setitem__(slice)

        keys = key_list[4:7]
        assert [key.meta().label for key in keys] == ["5", "6", "7"]
        key_list[4:7] = [Key(""), Key("")]
        assert [key.meta().label for key in keys] == ["", "", ""]

        # Test List.pop()

        last_key = key_list[9]
        assert last_key.meta().label == "10"
        key_list.pop()
        assert last_key.meta().label == ""

        # Test List.clear()

        keys = key_list[:]
        key_list.clear()
        assert all(key.meta().label == "" for key in keys)

    def test_paths(self) -> None:
        settings = ExampleSettings()

        assert settings.meta().path() == ""
        assert settings.inner_bunch.meta().path() == "inner_bunch"
        assert (
            settings.inner_bunch.bunch_key.meta().path()
            == "inner_bunch.bunch_key"
        )

        settings.bunch_list.appendOne()  # 1
        settings.bunch_list.appendOne()  # 2
        bunch = settings.bunch_list.appendOne()  # 3

        assert bunch.meta().path() == "bunch_list.3"



"SunsetSettings: a type-safe, extensible settings system with inheritance."

__project__ = "SunsetSettings"
__version__ = "0.6.0"
__author__ = "P. Varet"
__copyright__ = "2022-2024, P. Varet"

from . import exporter, serializers, sets

from .autosaver import AutoSaver
from .bunch import Bunch
from .key import Key
from .list import List
from .protocols import Serializable, Serializer
from .enum_serializer import (
    SerializableEnum,
    SerializableFlag,
)
from .settings import Settings, normalize
from .timer import PersistentTimer

# Backward compatibility: Bunch used to be called Bundle prior to version 0.4.0.
# Retain compatibility by keeping this name around for a few versions.
Bundle = Bunch

__all__ = [
    "AutoSaver",
    "Bunch",
    "Bundle",
    "Key",
    "List",
    "PersistentTimer",
    "Serializable",
    "SerializableEnum",
    "SerializableFlag",
    "Serializer",
    "Settings",
    "exporter",
    "sets",
    "normalize",
    "serializers",
]



import enum

from typing import Generic, Optional, TypeVar, cast

from sunset.enum_serializer import EnumSerializer
from sunset.protocols import Serializable, Serializer

_T = TypeVar("_T")
_Serializable = TypeVar("_Serializable", bound=Serializable)
_Castable = TypeVar("_Castable", int, float, str)


class StraightCastSerializer(Generic[_Castable]):
    _type: type[_Castable]

    def __init__(self, type_: type[_Castable]) -> None:
        self._type = type_

    def toStr(self, value: _Castable) -> str:
        return str(value)

    def fromStr(self, string: str) -> Optional[_Castable]:
        try:
            return self._type(string)
        except ValueError:
            return None


class BoolSerializer:
    def toStr(self, value: bool) -> str:
        return "true" if value else "false"

    def fromStr(self, string: str) -> Optional[bool]:
        string = string.strip().lower()
        if string in ("true", "yes", "y", "1"):
            return True
        if string in ("false", "no", "n", "0"):
            return False
        return None


class SerializableSerializer(Generic[_Serializable]):
    _type: type[_Serializable]

    def __init__(self, type_: type[_Serializable]) -> None:
        self._type = type_

    def toStr(self, value: _Serializable) -> str:
        return value.toStr()

    def fromStr(self, string: str) -> Optional[_Serializable]:
        return self._type.fromStr(string)


def lookup(type_: type[_T]) -> Optional[Serializer[_T]]:
    # Note the cast on the return values. It's unfortunate, but works around a
    # mypy limitation where it fails to recognize our serializers as rightful
    # implementation of the generic Serializer protocol.

    if issubclass(type_, Serializable):
        return cast(Serializer[_T], SerializableSerializer(type_))

    if type_ is bool:
        return cast(Serializer[_T], BoolSerializer())

    if issubclass(type_, enum.Enum):
        return cast(Serializer[_T], EnumSerializer(type_))

    # Note: these need to come after the Enum case, because Enum can be a
    # subclass of these.

    if (
        issubclass(type_, int)
        or issubclass(type_, float)
        or issubclass(type_, str)
    ):
        return cast(Serializer[_T], StraightCastSerializer(type_))

    return None



import enum

from sunset.serializers import lookup


class ExampleEnum(enum.Enum):
    ONE = 1
    TWO = 2


class ExampleFlag(enum.Flag):
    A = enum.auto()
    B = enum.auto()
    C = enum.auto()
    D = A | B | C
    OTHER_A = A


class IntEnumExample(enum.IntEnum):
    ONE = 1
    TWO = 2


class IntFlagExample(enum.IntFlag):
    A = enum.auto()
    B = enum.auto()


class TestSerializableEnum:
    def test_simple_enum(self) -> None:
        serializer = lookup(ExampleEnum)
        assert serializer is not None

        assert serializer.toStr(ExampleEnum.ONE) == "ONE"

        assert serializer.fromStr("") is None
        assert serializer.fromStr("TWO") == ExampleEnum.TWO
        assert serializer.fromStr("INVALID") is None
        assert serializer.fromStr("ONE|TWO") is None

    def test_flag_enum_tostr(self) -> None:
        ab = ExampleFlag.A | ExampleFlag.B
        ba = ExampleFlag.B | ExampleFlag.A
        abc = ab | ExampleFlag.C

        serializer = lookup(ExampleFlag)
        assert serializer is not None

        assert serializer.toStr(ExampleFlag.A) == "A"
        assert serializer.toStr(ExampleFlag.OTHER_A) == "A"
        assert serializer.toStr(ab) == "A|B"
        assert serializer.toStr(ba) == "A|B"
        assert serializer.toStr(ExampleFlag.D) == "D"
        assert serializer.toStr(abc) == "D"

    def test_flag_enum_fromstr(self) -> None:
        ab = ExampleFlag.A | ExampleFlag.B

        serializer = lookup(ExampleFlag)
        assert serializer is not None

        assert serializer.fromStr("") is None
        assert serializer.fromStr("A") == ExampleFlag.A
        assert serializer.fromStr("D") == ExampleFlag.D
        assert serializer.fromStr("OTHER_A") == ExampleFlag.A
        assert serializer.fromStr("OTHER_A") == ExampleFlag.OTHER_A
        assert serializer.fromStr("A|B") == ab
        assert serializer.fromStr("B|A") == ab
        assert serializer.fromStr("A | B") == ab
        assert serializer.fromStr("A|B|C") == ExampleFlag.D
        assert serializer.fromStr("A|B|INVALID") == ab
        assert serializer.fromStr("A|INVALID") == ExampleFlag.A
        assert serializer.fromStr("INVALID|A") == ExampleFlag.A
        assert serializer.fromStr("INVALID") is None

    def test_int_enum(self) -> None:
        serializer = lookup(IntEnumExample)
        assert serializer is not None

        assert serializer.toStr(IntEnumExample.ONE) == "ONE"
        assert serializer.fromStr("TWO") == IntEnumExample.TWO

    def test_int_flag(self) -> None:
        serializer = lookup(IntFlagExample)
        assert serializer is not None

        assert serializer.toStr(IntFlagExample.A | IntFlagExample.B) == "A|B"
        assert serializer.fromStr("A|B") == IntFlagExample.A | IntFlagExample.B

    def test_flexible_case_matching(self) -> None:
        class TestEnum(enum.Enum):
            TEST1 = enum.auto()
            test1 = enum.auto()
            TEST2 = enum.auto()

        serializer = lookup(TestEnum)
        assert serializer is not None
        assert serializer.fromStr("TEST1") == TestEnum.TEST1
        assert serializer.fromStr("test1") == TestEnum.test1
        assert serializer.fromStr("Test1") is None
        assert serializer.fromStr("TEST2") == TestEnum.TEST2
        assert serializer.fromStr("test2") == TestEnum.TEST2
        assert serializer.fromStr("Test2") == TestEnum.TEST2



import weakref

from dataclasses import dataclass
from types import GenericAlias
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

try:
    from typing import Self
except ImportError:
    # TODO: Remove once we deprecate support for Python 3.10.
    from typing_extensions import Self

from sunset.notifier import Notifier


_T = TypeVar("_T")


# pylint: disable=unnecessary-ellipsis
# The ellipses in the protocol definitions below are in fact necessary: they let
# the type checker know it's fine we're not returning values.


@runtime_checkable
class Serializable(Protocol):
    """
    A protocol to be implemented by a class in order to enable storing instances
    of that class in a Key.

    The two methods to be implemented are :meth:`toStr()` and (classmethod)
    :meth:`fromStr()`.
    """

    def toStr(self) -> str:
        """
        Returns a string representation of this instance that can be used by
        :meth:`fromStr()` to reconstruct a copy of this instance.
        """
        ...

    @classmethod
    def fromStr(cls: type[Self], string: str) -> Optional[Self]:
        """
        Takes a string that represents a serialized instance of this class, and
        returns a newly created instance that corresponds to that
        representation, or None if the string is not a valid serialized
        representation of an instance of this class.
        """
        ...


class Serializer(Generic[_T], Protocol):
    """
    A protocol that describes a way to serialize and deserialize an arbitrary
    type.

    SunsetSettings provides its own serializers for common types (int, float,
    bool, str, enum.Enum). In order to store an arbitrary type in a Key, users
    need to provide a serializer for that type when instantiating a Key. That
    serializer should be an implementation of this protocol.
    """

    def toStr(self, value: _T) -> str:
        """
        Returns a string representation of the given value, that can be used by
        :meth:`fromStr()` to reconstruct a copy of that value.
        """
        ...

    def fromStr(self, string: str) -> Optional[_T]:
        """
        Takes a string that represents a serialized instance of a value, and
        returns a newly created instance that corresponds to that string, or
        None if the string is not a valid representation of a value for this
        serializer's type.
        """
        ...


@runtime_checkable
class Inheriter(Protocol):
    def setParent(self, parent: Optional[Self]) -> None: ...

    def parent(self) -> Optional[Self]: ...

    def children(self) -> Iterator[Self]: ...


@runtime_checkable
class Dumpable(Protocol):
    def dumpFields(self) -> Iterator[tuple[str, Optional[str]]]: ...

    def restoreField(self, path: str, value: Optional[str]) -> bool: ...

    def isSet(self) -> bool: ...


@runtime_checkable
class UpdateNotifier(Protocol):
    _update_notifier: Notifier[["UpdateNotifier"]]


@runtime_checkable
class LoadedNotifier(Protocol):
    _loaded_notifier: Notifier[[]]

    def onLoadedCall(self, callback: Callable[[], Any]) -> None: ...


@runtime_checkable
class ItemTemplate(Protocol):
    def _typeHint(self) -> Union[type, GenericAlias]: ...

    def _newInstance(self) -> Self: ...


@runtime_checkable
class HasMetadata(Protocol):
    _PATH_SEPARATOR: str

    def meta(self) -> "Metadata": ...


@dataclass
class Metadata:
    container: Optional[weakref.ref[HasMetadata]] = None
    label: str = ""

    def clear(self) -> None:
        self.container = None
        self.label = ""

    def update(
        self,
        label: Optional[str] = None,
        container: Optional[HasMetadata] = None,
    ) -> None:
        if container is not None:
            self.container = weakref.ref(container)
        if label is not None:
            self.label = label

    def path(self) -> str:
        if self.container is None or (container := self.container()) is None:
            # Should be the empty string, in theory.
            return self.label

        path = container.meta().path()

        if not path:
            return self.label

        return path + container._PATH_SEPARATOR + self.label


@runtime_checkable
class Field(
    Dumpable,
    HasMetadata,
    Inheriter,
    ItemTemplate,
    UpdateNotifier,
    LoadedNotifier,
    Protocol,
): ...


class BaseField:
    _PATH_SEPARATOR: str = "."

    _metadata: Optional[Metadata] = None

    def skipOnSave(self) -> bool:
        """
        Internal.

        Returns whether this entity should be disregarded when saving these
        settings. For entities with an attribute name, it's equivalent to
        checking if the attribute is private (its name starts with an
        underscore). For entities with a section name, it's equivalent to
        checking if the section name is empty.

        Can be overridden in subclasses.

        Returns:
           A bool used internally by the settings saving logic.
        """

        return self.meta().label.startswith("_")

    def meta(self) -> Metadata:
        """
        Internal.

        Get the metadata object associated with this entity.

        Returns:
            A metadata object.
        """

        if self._metadata is None:
            self._metadata = Metadata()
        return self._metadata



import pytest

from pytest_mock import MockerFixture

from sunset import Bunch, Key, List, protocols


class ExampleBunch(Bunch):
    class InnerBunch(Bunch):
        b = Key(42)

    class Item(Bunch):
        c = Key("default c")

    a = Key("default a")
    inner_bunch = InnerBunch()
    list = List(Item(), order=List.PARENT_LAST)
    _private = Key("private")


class TestBunch:
    def test_protocol_implementation(self) -> None:
        bunch = ExampleBunch()
        assert isinstance(bunch, protocols.Field)

    def test_creation(self) -> None:
        bunch = ExampleBunch()
        bunch.list.appendOne()
        assert bunch.a.get() == "default a"
        assert bunch.inner_bunch.b.get() == 42
        assert bunch.list[0].c.get() == "default c"

    def test_uninstantiated_field_fails(self) -> None:
        class InnerBunch(Bunch):
            pass

        class FaultyBunch(Bunch):
            inner = InnerBunch

        with pytest.raises(TypeError):
            FaultyBunch()

    def test_inner_bunch_definition_is_fine(self) -> None:
        class FineBunch(Bunch):
            class InnerBunch(Bunch):
                pass

            inner = InnerBunch()

        FineBunch()

    def test_fields_cant_override_existing_attributes(self) -> None:
        # "__init__" is an attribute that happens to exist on the class.

        assert getattr(Bunch, "__init__", None) is not None

        class FaultyBunch(Bunch):
            __init__ = Key(default="test")  # type: ignore[assignment]

        with pytest.raises(TypeError):
            FaultyBunch()

    def test_inheritance(self) -> None:
        parent_bunch = ExampleBunch()
        child_bunch = ExampleBunch()

        assert child_bunch not in parent_bunch.children()
        assert child_bunch.parent() is None

        child_bunch.setParent(parent_bunch)
        assert child_bunch in parent_bunch.children()
        assert child_bunch.parent() is parent_bunch

        child_bunch.setParent(None)
        assert child_bunch not in parent_bunch.children()
        assert child_bunch.parent() is None

    def test_bunch_dataclass_mro(self) -> None:
        # Bunch instantiation is a tad tricky. Here we make sure that attributes
        # of a Bunch that are not SunsetSetting fields do work as expected.

        class ExampleBunchSubclass(Bunch):
            a = Key("test a")
            b = "random attribute"

            def test_a(self) -> str:
                return self.a.get()

        bunch = ExampleBunchSubclass()
        assert bunch.test_a() == "test a"
        assert bunch.b == "random attribute"

    def test_parenting(self) -> None:
        parent_bunch = ExampleBunch()
        child_bunch = ExampleBunch()
        child_bunch.setParent(parent_bunch)

        assert child_bunch.parent() is parent_bunch
        assert child_bunch in parent_bunch.children()

    def test_reparenting(self) -> None:
        bunch1 = ExampleBunch()
        bunch2 = ExampleBunch()
        child_bunch = ExampleBunch()

        assert child_bunch not in bunch1.children()
        assert child_bunch not in bunch2.children()
        assert child_bunch.parent() is None

        child_bunch.setParent(bunch1)
        assert child_bunch in bunch1.children()
        assert child_bunch not in bunch2.children()
        assert child_bunch.parent() is bunch1

        child_bunch.setParent(bunch2)
        assert child_bunch not in bunch1.children()
        assert child_bunch in bunch2.children()
        assert child_bunch.parent() is bunch2

        child_bunch.setParent(None)
        assert child_bunch not in bunch1.children()
        assert child_bunch not in bunch2.children()
        assert child_bunch.parent() is None

    def test_inheritance_propagation(self) -> None:
        parent_bunch = ExampleBunch()
        child_bunch = ExampleBunch()
        child_bunch.setParent(parent_bunch)

        assert child_bunch.a.parent() is parent_bunch.a
        assert child_bunch.inner_bunch.parent() is parent_bunch.inner_bunch
        assert child_bunch.inner_bunch.b.parent() is parent_bunch.inner_bunch.b
        assert child_bunch.list.parent() is parent_bunch.list

    def test_inheritance_values(self) -> None:
        parent_bunch = ExampleBunch()
        child_bunch = ExampleBunch()
        child_bunch.setParent(parent_bunch)

        parent_bunch.a.set("test parent")
        assert child_bunch.a.get() == "test parent"
        child_bunch.a.set("test child")
        assert child_bunch.a.get() == "test child"
        assert parent_bunch.a.get() == "test parent"

        parent_bunch.inner_bunch.b.set(101)
        assert child_bunch.inner_bunch.b.get() == 101
        child_bunch.inner_bunch.b.set(37)
        assert child_bunch.inner_bunch.b.get() == 37
        assert parent_bunch.inner_bunch.b.get() == 101

        parent_bunch.list.appendOne().c.set("test parent")
        assert [item.c.get() for item in parent_bunch.list.iter()] == [
            "test parent"
        ]
        assert [item.c.get() for item in child_bunch.list.iter()] == [
            "test parent"
        ]
        child_bunch.list.appendOne().c.set("test child")
        assert [item.c.get() for item in parent_bunch.list.iter()] == [
            "test parent"
        ]
        assert [item.c.get() for item in child_bunch.list.iter()] == [
            "test child",
            "test parent",
        ]
        del parent_bunch.list[0]
        assert [item.c.get() for item in parent_bunch.list.iter()] == []
        assert [item.c.get() for item in child_bunch.list.iter()] == [
            "test child"
        ]

    def test_dump_fields(self) -> None:
        bunch = ExampleBunch()
        assert list(bunch.dumpFields()) == []

        bunch.a.set("test dump a")
        bunch.inner_bunch.b.set(101)
        bunch.list.appendOne().c.set("test dump c 1")
        bunch.list.appendOne()
        bunch.list.appendOne().c.set("test dump c 3")
        bunch._private.set("test private")

        assert list(bunch.dumpFields()) == [
            ("a", "test dump a"),
            ("inner_bunch.b", "101"),
            ("list.1.c", "test dump c 1"),
            ("list.2", None),
            ("list.3.c", "test dump c 3"),
        ]

    def test_restore_field(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        # Test all flavors of valid paths. Also, restoring a field should not
        # trigger a callback.

        bunch = ExampleBunch()
        bunch.onUpdateCall(callback)

        assert bunch.a.get() == "default a"
        bunch.restoreField("a", "restore a")
        assert bunch.a.get() == "restore a"
        callback.assert_not_called()

        assert len(bunch.list) == 0
        bunch.restoreField("list.2.c", "restore c 2")
        bunch.restoreField("list.1.c", "restore c 1")
        assert len(bunch.list) == 2
        assert bunch.list[0].c.get() == "restore c 1"
        assert bunch.list[1].c.get() == "restore c 2"
        callback.assert_not_called()

        bunch.restoreField("list.1.c", "other restore c 1")
        assert len(bunch.list) == 2
        assert bunch.list[0].c.get() == "other restore c 1"
        assert bunch.list[1].c.get() == "restore c 2"
        callback.assert_not_called()

        assert bunch.inner_bunch.b.get() == 42
        bunch.restoreField("inner_bunch.b", "101")
        assert bunch.inner_bunch.b.get() == 101
        callback.assert_not_called()

        # Test invalid paths.

        other_bunch = ExampleBunch()
        other_bunch.onUpdateCall(callback)

        other_bunch.restoreField("x", "invalid path")
        assert not other_bunch.isSet()

        other_bunch.restoreField("a.a", "invalid path")
        assert not other_bunch.isSet()

        other_bunch.restoreField(".", "invalid path")
        assert not other_bunch.isSet()

        other_bunch.restoreField(".a", "invalid path")
        assert not other_bunch.isSet()

        other_bunch.restoreField("", "invalid path")
        assert not other_bunch.isSet()

    def test_persistence(self) -> None:
        # A Bunch does not keep a reference to its parent or children.

        bunch = ExampleBunch()
        level1 = ExampleBunch()
        level2 = ExampleBunch()
        level1.setParent(bunch)
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del bunch
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0

    def test_key_updated_notification(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        bunch = ExampleBunch()
        bunch.onUpdateCall(callback)

        child = ExampleBunch()
        child.setParent(bunch)

        bunch.a.set("test 1")
        callback.assert_called_once_with(bunch.a)
        callback.reset_mock()

        bunch.inner_bunch.b.set(123)
        callback.assert_called_once_with(bunch.inner_bunch.b)
        callback.reset_mock()

        bunch.list.appendOne()
        callback.assert_called_once_with(bunch.list)
        callback.reset_mock()

        bunch.list[0].c.set("test 2")
        callback.assert_called_once_with(bunch.list[0].c)
        callback.reset_mock()

        child.a.set("test 3")
        callback.assert_not_called()
        callback.reset_mock()

    def test_callback_type_is_flexible(self) -> None:
        bunch = ExampleBunch()

        class Dummy:
            pass

        def callback(_: ExampleBunch) -> Dummy:
            return Dummy()

        bunch.onUpdateCall(callback)

    def test_init_properly_called(self) -> None:
        called: list[Bunch] = []

        class TestBunch(Bunch):
            a = Key(default=0)

            def __init__(self) -> None:
                super().__init__()
                called.append(self)

        bunch = TestBunch()
        assert bunch in called
        assert bunch.a.get() == 0

    def test_missing_init_super_causes_attribute_error(self) -> None:
        class TestBunch(Bunch):
            a = Key(default=0)

            def __init__(self) -> None:
                # This incorrectly fails to call super().__init__().
                pass

        bunch = TestBunch()
        with pytest.raises(AttributeError):
            bunch.a



from typing import Iterator

from pytest_mock import MockerFixture

from sunset import Bunch, Key, List, protocols


class ExampleBunch(Bunch):
    test = Key("")


class TestList:
    def test_protocol_implementation(self) -> None:
        list_key: List[Key[int]] = List(Key(default=0))
        assert isinstance(list_key, protocols.Field)

    def test_add_pop(self) -> None:
        list_key: List[Key[str]] = List(Key(default=""))
        list_key.appendOne()

        assert len(list_key) == 1

        list_key[0].set("test")
        assert list_key[0].get() == "test"

        list_key.pop()

        assert len(list_key) == 0

    def test_inheritance(self) -> None:
        parent_list: List[Key[int]] = List(Key(default=0))
        child_list: List[Key[int]] = List(Key(default=0))

        assert child_list not in parent_list.children()
        assert child_list.parent() is None

        child_list.setParent(parent_list)
        assert child_list in parent_list.children()
        assert child_list.parent() is parent_list

        child_list.setParent(None)
        assert child_list not in parent_list.children()
        assert child_list.parent() is None

    def test_reparenting(self) -> None:
        list1: List[Key[int]] = List(Key(default=0))
        list2: List[Key[int]] = List(Key(default=0))
        child_list: List[Key[int]] = List(Key(default=0))

        assert child_list not in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is None

        child_list.setParent(list1)
        assert child_list in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is list1

        child_list.setParent(list2)
        assert child_list not in list1.children()
        assert child_list in list2.children()
        assert child_list.parent() is list2

        child_list.setParent(None)
        assert child_list not in list1.children()
        assert child_list not in list2.children()
        assert child_list.parent() is None

    def test_iter_inheritance(self) -> None:
        def flatten(keys: Iterator[Key[str]]) -> list[str]:
            return [key.get() for key in keys]

        parent: List[Key[str]] = List(Key(default=""))
        child_default: List[Key[str]] = List(Key(default=""))
        child_iter_no_parent: List[Key[str]] = List(
            Key(default=""), order=List.NO_PARENT
        )
        child_iter_parent_first: List[Key[str]] = List(
            Key(default=""), order=List.PARENT_FIRST
        )
        child_iter_parent_last: List[Key[str]] = List(
            Key(default=""), order=List.PARENT_LAST
        )

        parent.appendOne().set("parent")
        child_default.appendOne().set("child")
        child_iter_no_parent.appendOne().set("child")
        child_iter_parent_first.appendOne().set("child")
        child_iter_parent_last.appendOne().set("child")

        assert flatten(child_default.iter()) == ["child"]

        child_default.setParent(parent)
        child_iter_no_parent.setParent(parent)
        child_iter_parent_first.setParent(parent)
        child_iter_parent_last.setParent(parent)

        assert flatten(child_default.iter()) == ["child"]
        assert flatten(child_iter_no_parent.iter()) == ["child"]
        assert flatten(child_iter_parent_first.iter()) == ["parent", "child"]
        assert flatten(child_iter_parent_last.iter()) == ["child", "parent"]
        assert flatten(child_default.iter(order=List.NO_PARENT)) == ["child"]
        assert flatten(child_default.iter(order=List.PARENT_FIRST)) == [
            "parent",
            "child",
        ]
        assert flatten(child_default.iter(order=List.PARENT_LAST)) == [
            "child",
            "parent",
        ]

    def test_dump_fields(self) -> None:
        key_list = List(Key(""))
        key_list.appendOne().set("test 1")
        key_list.appendOne()
        key_list.appendOne().set("test 3")
        assert list(key_list.dumpFields()) == [
            ("1", "test 1"),
            ("2", None),
            ("3", "test 3"),
        ]

        bunch_list = List(ExampleBunch())
        bunch_list.appendOne().test.set("test 1")
        bunch_list.appendOne()
        bunch_list.appendOne().test.set("test 3")
        bunch_list.appendOne()
        assert list(bunch_list.dumpFields()) == [
            ("1.test", "test 1"),
            ("2", None),
            ("3.test", "test 3"),
            ("4", None),
        ]

        key_list = List(Key(""))
        key_list.appendOne()
        key_list.appendOne().set("")
        key_list.appendOne()
        assert list(key_list.dumpFields()) == [
            ("1", None),
            ("2", ""),
            ("3", None),
        ]

        class TestBunch(Bunch):
            key_list = List(Key(""))
            _private = List(Key(""))

        bunch = TestBunch()
        bunch.key_list.appendOne().set("test public")
        bunch._private.appendOne().set("test private")
        assert list(bunch.dumpFields()) == [
            ("key_list.1", "test public"),
        ]

    def test_restore_field(self, mocker: MockerFixture) -> None:
        callback = mocker.stub()

        # Test restoring one value. Also, restoring a field should not trigger a
        # callback.

        test_list1 = List(Key("default"))
        test_list1.onUpdateCall(callback)
        assert test_list1.restoreField("1", "test")
        assert len(test_list1) == 1
        assert test_list1[0].get() == "test"
        callback.assert_not_called()

        # Test overwriting a restored value.

        assert test_list1.restoreField("1", "other test")
        assert len(test_list1) == 1
        assert test_list1[0].get() == "other test"
        callback.assert_not_called()

        # Test restoring a value that requires extending the List.

        assert test_list1.restoreField("5", "extension test")
        assert len(test_list1) == 5
        assert test_list1[4].get() == "extension test"
        callback.assert_not_called()

        # Test different kinds of invalid restore paths.

        test_list2 = List(Key("default"))
        test_list2.onUpdateCall(callback)

        assert not test_list2.restoreField("", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        assert not test_list2.restoreField("invalid", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        assert not test_list2.restoreField("invalid.1", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        assert not test_list2.restoreField(".", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        assert not test_list2.restoreField("0", "test")
        assert len(test_list2) == 0
        callback.assert_not_called()

        # Test indirect restore path.

        class TestBunch(Bunch):
            str_list = List(Key("default"))

        bunch = TestBunch()
        bunch.str_list.onUpdateCall(callback)
        assert bunch.restoreField("str_list.1", "test")
        assert len(bunch.str_list) == 1
        assert bunch.str_list[0].get() == "test"
        callback.assert_not_called()

        # Test an invalid restore value.

        test_list3 = List(Key("default"))
        test_list3.onUpdateCall(callback)
        assert not test_list3.restoreField(".size", "invalid")
        assert len(test_list3) == 0
        callback.assert_not_called()

        # Test restoring a List with empty items.

        test_list4 = List(Key("default"))
        test_list4.onUpdateCall(callback)
        assert test_list4.restoreField("1", None)
        assert test_list4.restoreField("2", "")
        assert test_list4.restoreField("3", None)
        assert len(test_list4) == 3
        assert not test_list4[0].isSet()
        assert test_list4[1].isSet() and test_list4[1].get() == ""
        assert not test_list4[2].isSet()
        callback.assert_not_called()

    def test_persistence(self) -> None:
        # A List does not keep a reference to its parent or children.

        bunch_list: List[Key[int]] = List(Key(default=0))
        level1: List[Key[int]] = List(Key(default=0))
        level1.setParent(bunch_list)
        level2: List[Key[int]] = List(Key(default=0))
        level2.setParent(level1)

        assert level1.parent() is not None
        assert len(list(level1.children())) == 1

        del bunch_list
        del level2
        assert level1.parent() is None
        assert len(list(level1.children())) == 0

    def test_callback_type_is_flexible(self) -> None:
        key_list = List(Key(""))

        class Dummy:
            pass

        def callback(_: List[Key[str]]) -> Dummy:
            return Dummy()

        key_list.onUpdateCall(callback)

    def test_update_callback_called_on_list_contents_changed(
        self, mocker: MockerFixture
    ) -> None:
        callback = mocker.stub()

        key_list: List[Key[int]] = List(Key(default=0))

        key_list.onUpdateCall(callback)

        key_list.appendOne()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.append(Key(default=0))
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.insert(0, Key(default=0))
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list[0] = Key(default=0)
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list += [Key(default=0), Key(default=0), Key(default=0)]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.extend([Key(default=0), Key(default=0), Key(default=0)])
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list[2:4] = [
            Key(default=0),
            Key(default=0),
            Key(default=0),
        ]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key_list.pop()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        key = key_list[0]
        key_list.remove(key)
        assert key not in key_list
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        del key_list[0]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        assert len(key_list[1:3]) == 2
        del key_list[1:3]
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

        assert len(key_list) > 1
        key_list.clear()
        callback.assert_called_once_with(key_list)
        callback.reset_mock()

    def test_update_callback_called_on_contained_item_update(
        self, mocker: MockerFixture
    ) -> None:
        callback = mocker.stub()

        bunch_list: List[ExampleBunch] = List(ExampleBunch())
        bunch_list.onUpdateCall(callback)

        bunch1 = ExampleBunch()
        bunch_list.append(bunch1)
        callback.reset_mock()
        bunch1.test.set("test")
        callback.assert_called_once_with(bunch1.test)

        bunch2 = ExampleBunch()
        bunch_list.insert(0, bunch2)
        callback.reset_mock()
        bunch2.test.set("test")
        callback.assert_called_once_with(bunch2.test)

        bunch3 = ExampleBunch()
        bunch_list[0] = bunch3
        callback.reset_mock()
        bunch3.test.set("test")
        callback.assert_called_once_with(bunch3.test)

        bunch4 = ExampleBunch()
        bunch5 = ExampleBunch()
        bunch_list[1:2] = [bunch4, bunch5]
        callback.reset_mock()
        bunch4.test.set("test")
        callback.assert_called_once_with(bunch4.test)
        callback.reset_mock()
        bunch5.test.set("test")
        callback.assert_called_once_with(bunch5.test)

        bunch6 = ExampleBunch()
        bunch7 = ExampleBunch()
        bunch_list += [bunch6, bunch7]
        callback.reset_mock()
        bunch6.test.set("test")
        callback.assert_called_once_with(bunch6.test)
        callback.reset_mock()
        bunch7.test.set("test")
        callback.assert_called_once_with(bunch7.test)

        bunch8 = ExampleBunch()
        bunch9 = ExampleBunch()
        bunch_list.extend([bunch8, bunch9])
        callback.reset_mock()
        bunch8.test.set("test")
        callback.assert_called_once_with(bunch8.test)
        callback.reset_mock()
        bunch9.test.set("test")
        callback.assert_called_once_with(bunch9.test)

    def test_update_callback_not_called_for_removed_items(
        self, mocker: MockerFixture
    ) -> None:
        bunch_list: List[ExampleBunch] = List(ExampleBunch())

        bunch1 = ExampleBunch()
        bunch2 = ExampleBunch()

        callback = mocker.stub()
        bunch_list.onUpdateCall(callback)

        bunch1.test.clear()
        bunch_list[:] = [bunch1]
        del bunch_list[0]

        callback.reset_mock()
        bunch1.test.set("test")
        callback.assert_not_called()

        bunch1.test.clear()
        bunch2.test.clear()
        bunch_list[:] = [bunch1, bunch2]
        del bunch_list[0:2]

        callback.reset_mock()
        bunch1.test.set("test")
        bunch2.test.set("test")
        callback.assert_not_called()

        bunch1.test.clear()
        bunch2.test.clear()
        bunch_list[:] = [bunch1, bunch2]
        bunch_list.clear()

        callback.reset_mock()
        bunch1.test.set("test")
        bunch2.test.set("test")
        callback.assert_not_called()

        bunch1.test.clear()
        bunch_list[:] = [bunch1]
        bunch_list.pop()

        callback.reset_mock()
        bunch1.test.set("test")
        callback.assert_not_called()

        bunch1.test.clear()
        bunch_list[:] = [bunch1]
        bunch_list.remove(bunch1)
        assert bunch1 not in bunch_list

        callback.reset_mock()
        bunch1.test.set("test")
        callback.assert_not_called()

    def test_repr(self) -> None:
        parent = List(Key(default=0))
        list_iter_no_parent = List(Key(default=0), order=List.NO_PARENT)
        list_iter_parent_first = List(Key(default=0), order=List.PARENT_FIRST)
        list_iter_parent_last = List(Key(default=0), order=List.PARENT_LAST)

        list_iter_no_parent.setParent(parent)
        list_iter_parent_first.setParent(parent)
        list_iter_parent_last.setParent(parent)

        list_iter_no_parent.appendOne().set(1)
        list_iter_parent_first.appendOne().set(1)
        list_iter_parent_last.appendOne().set(1)

        assert repr(list_iter_no_parent) == "[<Key[int]:1>]"
        assert repr(list_iter_parent_first) == "[<parent>,<Key[int]:1>]"
        assert repr(list_iter_parent_last) == "[<Key[int]:1>,<parent>]"



import threading
import time
import typing

from sunset import Key, List, Settings


_P = typing.ParamSpec("_P")

_THREAD_COUNT = 16
_DURATION = 0.1  # seconds
_ATTEMPTS = 5


def run_threaded(
    func: typing.Callable[typing.Concatenate[int, _P], None],
    thread_count: int = _THREAD_COUNT,
    duration: float = _DURATION,
    *args: _P.args,
    **kwargs: _P.kwargs,
) -> None:
    exceptions: list[Exception] = []
    deadline: float = time.monotonic() + duration
    barrier = threading.Barrier(parties=thread_count)

    def executor(thread_id: int) -> None:
        barrier.wait()
        try:
            while time.monotonic() < deadline:
                func(thread_id, *args, **kwargs)
        except Exception as e:
            exceptions.append(e)

    threads = [
        threading.Thread(target=executor, args=[i]) for i in range(thread_count)
    ]

    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join()

    if exceptions:
        raise exceptions[0]


class TestKeyConcurrency:
    def test_set_clear(self) -> None:
        str_key = Key("default value")

        def clear_and_set_key(thread_id: int) -> None:
            str_key.clear()
            str_key.set(str(thread_id))

        for _ in range(_ATTEMPTS):
            run_threaded(clear_and_set_key)

            assert str_key.isSet()
            assert str_key.get() != "default value"

        def set_and_clear_key(thread_id: int) -> None:
            str_key.set(str(thread_id))
            str_key.clear()

        for _ in range(_ATTEMPTS):
            run_threaded(set_and_clear_key)

            assert not str_key.isSet()
            assert str_key.get() == "default value"

    def test_update_value(self) -> None:
        count = [0]
        int_key = Key(0)
        count_lock = threading.Lock()

        def updater(value: int) -> int:
            return value + 1

        def update_key(_unused: int) -> None:
            int_key.updateValue(updater)
            with count_lock:
                count[0] += 1

        for _ in range(_ATTEMPTS):
            run_threaded(update_key)

        assert int_key.get() == count[0]

    def test_parenting(self) -> None:
        parent_key = Key("")
        parent_key.set("parent")
        str_key = Key("")

        def set_parent(thread_id: int) -> None:
            str_key.setParent(None)
            str_key.setParent(parent_key)

        for _ in range(_ATTEMPTS):
            run_threaded(set_parent)

            assert str_key.parent() is parent_key
            assert str_key.get() == "parent"


class TestListConcurrency:
    def test_append_pop(self) -> None:
        _start_items: int = 10

        key_list = List(Key(""))
        for i in range(_start_items):
            key_list.append(Key(f"default {i}"))

        def update_list(thread_id: int, key_list: List[Key[str]]) -> None:
            value = str(thread_id)
            if (fashion := thread_id % 4) == 0:
                key_list.append(Key(value))
            elif fashion == 1:
                key_list += [Key(value)]
            elif fashion == 2:
                key_list[0:0] = [Key(value)]
            elif fashion == 3:
                key_list.insert(0, Key(value))

            if (fashion := (thread_id << 2) % 2) == 0:
                key_list.pop()
            elif fashion == 1:
                del key_list[0]

            assert len(key_list) >= _start_items

        for _ in range(_ATTEMPTS):
            run_threaded(update_list, key_list=key_list)

            assert len(key_list) == _start_items
            for i, item in enumerate(key_list):
                assert item in key_list
                expected_label = key_list._labelForIndex(i)
                assert item.meta().label == expected_label


class TestSettingsConcurrency:
    def test_settings_names(self) -> None:
        class TestSettings(Settings):
            pass

        settings = TestSettings()
        sections = [settings.newSection(str(i)) for i in range(16)]

        def rename_section(
            thread_id: int, sections: list[TestSettings]
        ) -> None:
            section = sections[thread_id % len(sections)]
            section.setSectionName("test")
            section.setSectionName("section")

        for _ in range(_ATTEMPTS):
            run_threaded(rename_section, sections=sections)

            assert len(list(settings.sections())) == len(sections)
            assert len(
                set(section.sectionName() for section in settings.sections())
            ) == len(sections)



import threading

from functools import wraps
from typing import Callable, Concatenate, ParamSpec, TypeVar


_Self = TypeVar("_Self", bound="Lockable")
_R = TypeVar("_R")
_P = ParamSpec("_P")


class Lockable:
    """
    Internal.

    A helper mixin that provides a context manager to lock a class method's
    against one another.
    """

    _lock: threading.RLock

    def __init__(self) -> None:
        super().__init__()

        self._lock = threading.RLock()

    @staticmethod
    def with_lock(
        method: Callable[Concatenate[_Self, _P], _R]
    ) -> Callable[Concatenate[_Self, _P], _R]:
        @wraps(method)
        def locked_method(
            self: _Self, *args: _P.args, **kwargs: _P.kwargs
        ) -> _R:
            with self._lock:
                return method(self, *args, **kwargs)

        return locked_method



import logging
import pathlib

from typing import (
    Any,
    Callable,
    IO,
    Iterator,
    MutableSet,
    Optional,
    Union,
)

try:
    from typing import Self
except ImportError:
    # TODO: Remove once we deprecate support for Python 3.10.
    from typing_extensions import Self

from sunset.autosaver import AutoSaver
from sunset.bunch import Bunch
from sunset.exporter import normalize, load_from_file, save_to_file
from sunset.lockable import Lockable
from sunset.sets import NonHashableSet


_MAIN = "main"


class Settings(Bunch, Lockable):
    """
    A collection of keys that can be saved to and loaded from text, and supports
    subsections.

    Under the hood, a Settings class is a dataclass, and can be used in the same
    manner, i.e. by defining attributes directly on the class itself.

    Settings instances support subsections: calling the :meth:`newSection()`
    method on an instance creates a subsection of that instance. This subsection
    holds the same keys, with independant values. If a key of the subsection
    does not have a value, its value will be looked up on its parent section
    instead. The hierarchy of sections can be arbitrarily deep.

    When saving a Settings instance, its subsections are saved with it under a
    distinct heading for each, provided they have a name. A section is given a
    name by passing it to the :meth:`newSection()` method, or by using the
    :meth:`setSectionName()` method on the new section after creation.

    The name of each section is used to construct the heading it is saved under.
    The top-level Settings instance is saved under the `[main]` heading by
    default.

    Anonymous (unnamed) sections do not get saved.

    Example:

    >>> from sunset import Key, Settings
    >>> class AnimalSettings(Settings):
    ...     hearts: Key[int] = Key(default=0)
    ...     legs: Key[int]   = Key(default=0)
    ...     wings: Key[int]  = Key(default=0)
    ...     fur: Key[bool]   = Key(default=False)
    >>> animals = AnimalSettings()
    >>> animals.hearts.set(1)
    True
    >>> mammals = animals.newSection(name="mammals")
    >>> mammals.fur.set(True)
    True
    >>> mammals.legs.set(4)
    True
    >>> humans = mammals.newSection(name="humans")
    >>> humans.legs.set(2)
    True
    >>> humans.fur.set(False)
    True
    >>> birds = animals.newSection(name="birds")
    >>> birds.legs.set(2)
    True
    >>> birds.wings.set(2)
    True
    >>> aliens = animals.newSection()  # No name given!
    >>> aliens.hearts.set(2)
    True
    >>> aliens.legs.set(7)
    True
    >>> print(mammals.hearts.get())
    1
    >>> print(mammals.legs.get())
    4
    >>> print(mammals.wings.get())
    0
    >>> print(mammals.fur.get())
    True
    >>> print(birds.hearts.get())
    1
    >>> print(birds.legs.get())
    2
    >>> print(birds.wings.get())
    2
    >>> print(birds.fur.get())
    False
    >>> print(humans.hearts.get())
    1
    >>> print(humans.legs.get())
    2
    >>> print(humans.wings.get())
    0
    >>> print(humans.fur.get())
    False
    >>> print(aliens.hearts.get())
    2
    >>> print(aliens.legs.get())
    7
    >>> print(aliens.wings.get())
    0
    >>> print(aliens.fur.get())
    False
    >>> import io
    >>> txt = io.StringIO()
    >>> animals.save(txt)
    >>> print(txt.getvalue(), end="")
    [main]
    hearts = 1
    [birds]
    legs = 2
    wings = 2
    [mammals]
    fur = true
    legs = 4
    [mammals/humans]
    fur = false
    legs = 2
    """

    MAIN: str = _MAIN

    _SECTION_SEPARATOR = "/"

    _section_name: str = ""
    _children_set: MutableSet[Bunch]
    _autosaver: Optional[AutoSaver] = None
    _autosaver_class: type[AutoSaver]

    def __post_init__(self) -> None:
        super().__post_init__()

        # Note that this overrides the similarly named attribute from the parent
        # class. In the parent class, the set does not keep references to its
        # items; in this class, it does.

        self._children_set = NonHashableSet()
        self._autosaver_class = AutoSaver

    @Lockable.with_lock
    def newSection(self, name: str = "") -> Self:
        """
        Creates and returns a new instance of this class. Each key of the new
        instance will inherit from the key of the same name on the parent
        instance.

        When saving Settings with the :meth:`save()` method, each section's name
        is used to generate the heading under which that section is saved. If
        the new section is created without a name, it will be skipped when
        saving. A name can still be given to a section after creation with the
        :meth:`setSectionName()` method.

        If this Settings instance already has a section with the given name, the
        new section will be created with a unique name generated by appending a
        numbered suffix to that name.

        Args:
            name: The name that will be used to generate a heading for this
                section when saving it to text. The given name will be
                normalized to lowercase alphanumeric characters.

        Returns:
            An instance of the same type as self.
        """

        new = self._newInstance()
        new.setSectionName(name)

        # Note that this will trigger an update notification.

        new.setParent(self)

        return new

    @Lockable.with_lock
    def getOrCreateSection(self, name: str) -> Self:
        """
        Finds and returns the section of these Settings with the given name if
        it exists, and creates it if it doesn't.

        If the given name is empty, this is equivalent to calling
        :meth:`newSection()` instead.

        Args:
            name: The name that will be used to generate a heading for this
                section when saving it to text. The given name will be
                normalized to lowercase alphanumeric characters.

        Returns:
            An instance of the same type as self.
        """

        return (
            section
            if (section := self.getSection(name)) is not None
            else self.newSection(name=name)
        )

    def getSection(self, name: str) -> Optional[Self]:
        """
        Finds and returns a section of this instance with the given name, if it
        exists, else None.

        Args:
            name: The name of the section to return.

        Returns:
            An instance of the same type as self, or None.
        """

        norm = normalize(name)
        if not norm:
            return None

        for section in self.sections():
            if norm == section.sectionName():
                return section

        return None

    def sections(self) -> Iterator[Self]:
        """
        Returns an iterator over the subsections of this Settings instance. Note
        that the subsections are only looked up one level deep, that is to say,
        no recursing into the section hierarchy occurs.

        Returns:
            An iterator over Settings instances of the same type as this one.
        """

        yield from sorted(self.children())

    def setSectionName(self, name: str) -> str:
        """
        Sets the unique name under which this Settings instance will be
        persisted by the :meth:`save()` method. The given name will be
        normalized to lowercase, without space or punctuation.

        This name is guaranteed to be unique. If the given name is already used
        by a section of the same Settings instance, then a numbered suffix is
        generated to make this one's name unique.

        If the given name is empty, these settings will be skipped when saving.

        The toplevel Settings instance is named "main" by default. It cannot be
        unnamed; it will revert to its default name instead.

        Args:
            name: The name that will be used to generate a heading for these
                settings when saving them to text.

        Returns:
            The given name, normalized, and made unique if needed.

        Example:

        >>> from sunset import Settings
        >>> class TestSettings(Settings):
        ...     pass
        >>> parent = TestSettings()
        >>> section1 = parent.newSection()
        >>> section2 = parent.newSection()
        >>> section3 = parent.newSection()
        >>> section1.setSectionName("  T ' e ? S / t")
        'test'
        >>> section2.setSectionName("TEST")
        'test_1'
        >>> section3.setSectionName("test")
        'test_2'
        >>> # This should not change this section's name.
        >>> section3.setSectionName("test")
        'test_2'
        """

        name = normalize(name)

        if (parent := self.parent()) is None:
            if name != self._section_name:
                self._section_name = name
                self._update_notifier.trigger(self)

        else:
            # Note that this triggers a notification if the unique name is
            # different from the current name.

            parent._setUniqueNameForSection(name, self)

        return self.sectionName()

    @Lockable.with_lock
    def _setUniqueNameForSection(self, name: str, section: Self) -> None:
        candidate = name = normalize(name)

        if candidate:
            other_names = set(
                s.sectionName() for s in self.children() if s is not section
            )

            i = 0
            while candidate in other_names:
                i += 1
                candidate = f"{name}_{i}"

        if candidate != section._section_name:
            section._section_name = candidate
            section._update_notifier.trigger(section)

    def sectionName(self) -> str:
        """
        Returns the current name of this Settings instance. This name will be
        used to generate the heading under which this Settings instance will be
        persisted by the :meth:`save()` method.

        Returns:
            The name of this Settings instance.
        """
        if name := self._section_name:
            return name

        return name if self.parent() is not None else self.MAIN

    @Lockable.with_lock
    def setParent(self, parent: Optional[Self]) -> None:  # type: ignore
        """
        Makes the given Settings instance the parent of this one. If None,
        remove this instance's parent, if any.

        All the Key, List and Bunch fields defined on this instance will be
        recursively reparented to the corresponding Key / List / Bunch field on
        the given parent.

        This method is for internal purposes and you will typically not need to
        call it directly.

        Args:
            parent: Either a Settings instance that will become this instance's
                parent, or None. The parent Settings instance must have the same
                type as this instance.
        """

        if parent is (previous_parent := self.parent()):
            return

        # Ensure that this section's name is unique in its parent.

        if parent is not None:
            # May trigger an update notification if the name is changed.

            parent._setUniqueNameForSection(self._section_name, self)
            self._update_notifier.add(parent._update_notifier.trigger)

        super().setParent(parent)  # type:ignore

        if parent is not None:
            parent._update_notifier.trigger(parent)

        if previous_parent is not None:
            previous_parent._update_notifier.trigger(previous_parent)
            self._update_notifier.discard(
                previous_parent._update_notifier.trigger
            )

    # Not actually useless. This lets us override the docstring with
    # Settings-specific info.
    # pylint: disable-next=useless-parent-delegation
    def onUpdateCall(self, callback: Callable[[Any], Any]) -> None:
        """
        Adds a callback to be called whenever this Settings instance is updated.
        A Settings instance is considered updated when any of its fields is
        updated, or when its name is updated.

        The callback will be called with as its argument whichever field was
        just updated.

        Args:
            callback: A callable that will be called with one argument of type
                :class:`~sunset.List`, :class:`~sunset.Bunch` or
                :class:`~sunset.Key`.

        Note:
            This method does not increase the reference count of the given
            callback.
        """

        super().onUpdateCall(callback)

    def skipOnSave(self) -> bool:
        return self.sectionName() == ""

    def dumpFields(self) -> Iterator[tuple[str, Optional[str]]]:
        """
        Internal.
        """

        if not self.skipOnSave():
            # Ensure the section is dumped event if empty. Dumping an empty
            # section is valid.

            label = (self.sectionName() or "?") + self._SECTION_SEPARATOR

            if not self.isSet():
                yield label, None
            else:
                yield from (
                    (label + path, item) for path, item in super().dumpFields()
                )

            for section in self.sections():
                yield from (
                    (label + path, item) for path, item in section.dumpFields()
                )

    def restoreField(self, path: str, value: Optional[str]) -> bool:
        """
        Internal.
        """

        if self._SECTION_SEPARATOR not in path:
            return False

        section_name, path = path.split(self._SECTION_SEPARATOR, 1)
        if self.sectionName() != section_name:
            return False

        with self._update_notifier.inhibit():
            if self._SECTION_SEPARATOR in path:
                subsection_name, _ = path.split(self._SECTION_SEPARATOR, 1)
                if subsection_name:
                    section = self.getOrCreateSection(subsection_name)
                    return section.restoreField(path, value)

            else:
                return super().restoreField(path, value)

        return False

    def save(self, file: IO[str], blanklines: bool = False) -> None:
        """
        Writes the contents of this Settings instance and its subsections in
        text form to the given file object.

        Args:
            file: A text file object where to save this Settings instance.

            blanklines: Whether to add a blank line before section headings.
        """

        if self.skipOnSave():
            # This is an anonymous instance, actually. There is therefore
            # nothing to save.

            return

        save_to_file(file, self.dumpFields(), blanklines=blanklines)

    def load(self, file: IO[str]) -> None:
        """
        Loads settings from the given text file object.

        If the given file contains lines that don't make sense -- for instance,
        if the line is garbage, or refers to a key that does not exist in this
        Settings class, or it exists but with an incompatible type -- then the
        faulty line is skipped silently.

        If the text file object contains multiple headings, those headings will
        be used to create subsections with the corresponding names.

        Loading settings from a file does not reset the existing settings or
        sections. Which means you can split your configuration into multiple
        files if needed and load each file in turn to reconstruct the full
        settings.

        Args:
            file: A text file open in reading mode.
        """

        for path, dump in load_from_file(file, self.sectionName()):
            self.restoreField(path, dump)

        self._loaded_notifier.trigger()

    def setAutosaverClass(self, class_: type[AutoSaver]) -> None:
        """
        Internal.
        """
        self._autosaver_class = class_

    def autosave(
        self,
        path: Union[str, pathlib.Path],
        save_on_update: bool = True,
        save_delay: int = 0,
        logger: Optional[logging.Logger] = None,
    ) -> AutoSaver:
        """
        Returns a context manager that loads these Settings from the given file
        path on instantiation and saves them on exit.

        By default, it will also automatically save the settings whenever they
        are updated from inside the application. Optionally, the updates can be
        batched over a given delay before being saved.

        See the documentation of :class:`~sunset.AutoSaver` for the details.

        Args:
            path: The full path to the file to load the settings from and save
                them to. If this file does not exist yet, it will be created
                when saving for the first time.

            save_on_update: Whether to save the settings when they
                are updated in any way. Default: True.

            save_delay: How long to wait, in seconds, before actually saving the
                settings when `save_on_update` is True and an update occurs.
                Setting this to a few seconds will batch updates for that long
                before triggering a save. If set to 0, the save is triggered
                immediately. Default: 0.

            logger: A logger instance that will be used to log OS errors, if
                any, while loading or saving settings. If none is given, the
                default root logger will be used.

        Returns:
            An AutoSaver context manager.
        """

        # Keep a reference to the AutoSaver instance, so that its lifetime is
        # bound to that of this Settings instance.

        self._autosaver = self._autosaver_class(
            self,
            path,
            save_on_update=save_on_update,
            save_delay=save_delay,
            logger=logger,
        )
        return self._autosaver

    def __lt__(self, other: Self) -> bool:
        # Giving sections an order lets us easily sort them when dumping.

        return self.sectionName() < other.sectionName()



import threading

from unittest.mock import MagicMock

from pytest import MonkeyPatch

from sunset import PersistentTimer


class TestPersistentTimer:
    def test_timer_no_delay(self, monkeypatch: MonkeyPatch) -> None:
        mock_cls = MagicMock(threading.Timer, autospec=True)
        mock = mock_cls.return_value
        monkeypatch.setattr(threading, "Timer", mock_cls)

        def test() -> None:
            pass

        timer = PersistentTimer(test)

        assert timer._timer is None  # type: ignore

        timer.start(0.0)

        mock_cls.assert_not_called()
        mock.start.assert_not_called()

        assert timer._timer is None  # type: ignore

        timer.cancel()

        assert timer._timer is None  # type: ignore

    def test_timer_with_delay(self, monkeypatch: MonkeyPatch) -> None:
        mock_cls = MagicMock(threading.Timer, autospec=True)
        mock = mock_cls.return_value
        monkeypatch.setattr(threading, "Timer", mock_cls)

        def test() -> None:
            pass

        timer = PersistentTimer(test)

        assert timer._timer is None  # type: ignore

        timer.start(1.0)

        mock_cls.assert_called_once_with(1.0, timer._timeout)  # type: ignore
        assert timer._timer is mock  # type: ignore
        mock.start.assert_called_once()  # type: ignore

        timer.cancel()

        assert timer._timer is None  # type: ignore
